/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamproducerqueue.h"
#include "gdamutil.h"
#include "gdamsongdatabase.h"
#include <gtk/gtksignal.h>
#include <glib.h>
#include <string.h>

enum {
	SET_FILENAME,
	SET_QUEUE_LIST,
	LAST_SIGNAL
};

static guint producer_queue_signals[LAST_SIGNAL];


static void     gdam_producer_queue_real_add_filename
					     (GdamProducerQueue* queue,
					      const char*    	 filename);
static void gdam_producer_queue_real_set_filename(GdamProducerQueue* queue,
                                                  const char*      filename);
static void     gdam_producer_queue_real_skip_songs
					     (GdamProducerQueue* queue,
					      int    		 num_songs);
static void gdam_producer_queue_real_skip_to (GdamProducerQueue* queue,
                                             int      	         song_num);
static void gdam_producer_queue_real_set_queue_list(GdamProducerQueue* queue,
						    gpointer 	       data);
static void     gdam_producer_queue_destroy  (GtkObject*         object);
static void     gdam_producer_queue_real_play  (GdamProducer*    producer,
					        GdamPlayInfo*    play_info);
static void mp3_ended_callback(gpointer user_data);


static GtkObjectClass* parent_class;
static void gdam_producer_queue_class_init(GdamProducerQueueClass* klass);

static void gdam_producer_queue_init(GdamProducerQueue* queue)
{
	queue->fade_envelope = 64;
}

GtkType gdam_producer_queue_get_type()
{
	static GtkType queue_type = 0;
	if (!queue_type) {
		static const GtkTypeInfo queue_info = {
			"GdamProducerQueue",
			sizeof(GdamProducerQueue),
			sizeof(GdamProducerQueueClass),
			(GtkClassInitFunc) gdam_producer_queue_class_init,
			(GtkObjectInitFunc) gdam_producer_queue_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		queue_type = gtk_type_unique(GDAM_TYPE_PRODUCER, 
					     &queue_info);
	}
	return queue_type;
}

static void queue_eof_connect_success(int eof_sub_id, gpointer user_data)
{
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(user_data);
	queue->queued_eof_sub_id = eof_sub_id;
	gdam_channel_register_handler(
		GDAM_MODEL(queue)->channel,
		gdam_channel_simple_handler_new(eof_sub_id,
						TRUE,
						mp3_ended_callback,
						queue)
	);
}

static void queued_source_success_callback(int new_id, gpointer user_data){
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(user_data);
	queue->queued_id = new_id;
	gtk_object_unref(GTK_OBJECT(queue));
}

static void queued_source_failure_callback(gpointer user_data){
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(user_data);
	queue->queued_id = 0;
	gtk_object_unref(GTK_OBJECT(queue));
}


/*  Determines which song should come after then current one, 
 *  and sets it up on the server side. */
static void queue_refresh_queued_song(GdamProducerQueue* queue)
{
	int queued_source_id;
	int source_id_array[2];
	int cur_idx;
	GdamRequest *request, *srequest;
	GdamRequestBuilder* request_builder;
	GdamArg arg;
	GdamSongInfo* queued_song;

	if (!queue->current_song)
		return;

	cur_idx = g_list_index(queue->songinfo_list, queue->current_song);
	queued_song = g_list_nth_data(queue->songinfo_list, cur_idx+1);

	if (queued_song == queue->queued_song)
		return;
	queue->queued_song = queued_song;

	/* The rest of the function updates the sources if playing */

	if (!GDAM_MODEL(queue)->source_id)
		return;

	request_builder = gdam_request_builder_new();

	if (queue->queued_song){
		request = gdam_request_create_source(
				"GdamSeekableMp3",
				queued_source_success_callback,
				queued_source_failure_callback,
				queue,
				3,
				"filename",
				GDAM_ARG_TYPE_STRING,
				gdam_song_info_get_filename(queue->queued_song),
				"position",
				GDAM_ARG_TYPE_INT64,
				(gint64)0,
				"buffer_size",
				GDAM_ARG_TYPE_INT,
				(gint)queue->buffer_size);

		queued_source_id = gdam_request_builder_append_create(
						request_builder,
						request);
		gdam_request_builder_append_generic(
			request_builder,
			gdam_request_subscribe_source_ended(
				queued_source_id,
				queue_eof_connect_success,
				NULL,
				queue)
		);
		source_id_array[1] = queued_source_id;
		gdam_debug("Queue: refreshing with %s",
			       gdam_song_info_get_filename(queue->queued_song));
	}
	else{
		source_id_array[1] = 0;
		gdam_debug("Queue: refreshing with no song.");
	}

	source_id_array[0] = queue->current_id;

	arg.name = "queue";
	arg.type  = GDAM_ARG_TYPE_INT_ARRAY;
	GDAM_ARG_INT_ARRAY(&arg).length = 2;
	GDAM_ARG_INT_ARRAY(&arg).array = source_id_array;

	srequest = gdam_request_set_args(
			GDAM_MODEL(queue)->source_id,
			1,
			&arg,
			NULL,
			NULL,
			NULL);

	gdam_request_builder_append_generic(request_builder, srequest);

	 /* Ref once for queued mp3 request
	  */
	gtk_object_ref(GTK_OBJECT(queue));

	gdam_channel_append_request(
			GDAM_MODEL(queue)->channel, 
			gdam_request_builder_package(request_builder));
			
}

/* Delete a range of songs.  counts from 1. */
void gdam_producer_queue_delete_range(GdamProducerQueue* queue,
				      guint lower, guint upper)
{
	int n;    /*  n varies from lower to upper, on the same scale.  */
	int to_play = 0;  /* also counted from 1, for now. */
	gboolean delete_need_refresh = FALSE;
	GList* songinfo_list = g_list_copy(queue->songinfo_list);
	g_return_if_fail(lower <= upper);

	gdam_debug("queue delete_range: %d %d", lower, upper);

	for (n=upper;n>=lower;n--){
		gpointer data;

		if (n > g_list_length(songinfo_list))
			continue;
		data = g_list_nth_data(songinfo_list, n-1);
		if (data){
			if (data == queue->current_song){
				if (n == g_list_length(songinfo_list)){
					to_play = n - 1;
				}
				else{
					to_play = n + 1;
				}
				delete_need_refresh = TRUE;
			}
			songinfo_list = g_list_remove(songinfo_list, data);
		}
	}
	/* TODO: set_queue_list does (some) refreshing... */
	gdam_producer_queue_set_queue_list(queue, songinfo_list);

	if (delete_need_refresh){
		if (to_play > 0)
			gdam_producer_queue_skip_to(queue, to_play-1);
		else
			gdam_producer_stop(GDAM_PRODUCER(queue));
	}
}

/* Move a range of songs.  counts from 1. */
void gdam_producer_queue_move_range(GdamProducerQueue* queue,
				      guint lower, guint upper, guint dest)
{
	int n;    /*  n varies from lower to upper, on the same scale.  */
	GList* songinfo_list = g_list_copy(queue->songinfo_list);
	  /* TODO: what if dest is  zero? */
	GList* target_link = g_list_nth(songinfo_list, dest-1);
	GList* mobile_list = NULL;

	gdam_debug("queue move_range: %d %d to %d", lower, upper, dest);

	g_return_if_fail(lower <= upper);
	if (dest <= lower && dest >= upper)
			return;

	for (n=upper;n>=lower;n--){
		gpointer data;

		if (n > g_list_length(songinfo_list))
			continue;
		data = g_list_nth_data(songinfo_list, n-1);
		if (data){
			songinfo_list = g_list_remove(songinfo_list, 
						      data);
			mobile_list = g_list_prepend(mobile_list, data);
		}
	}
	/*  what if dest is NULL?  concat lists.  */
	if (target_link){
		int tar = g_list_position(songinfo_list, target_link);
		while (mobile_list){
			gpointer data = mobile_list->data;
			mobile_list = g_list_remove(mobile_list, data);
			songinfo_list = g_list_insert(songinfo_list, 
						      data,
						      tar++);
		}
	}
	else {
		g_list_concat(songinfo_list, mobile_list);
	}
	gdam_producer_queue_set_queue_list(queue, songinfo_list);
}


static void gdam_producer_queue_real_add_filename(GdamProducerQueue* queue,
                                                  const char*      filename)
{
	GdamSongDatabase* database = gdam_song_database_system();
	GdamSongInfo* info = NULL;
	GList* songinfo_list = g_list_copy(queue->songinfo_list);

	gdam_debug("queue_real_add_filename");

	if (database)
		info = gdam_song_database_find(database, filename);

	if (!info) {
		info = gdam_song_info_new(filename, 0, 0);
	}
	else
		info = gdam_song_info_copy(info);

	songinfo_list = g_list_append(songinfo_list, info);

	gdam_producer_queue_set_queue_list(queue, songinfo_list);

	if (g_list_length(songinfo_list) == 1){
		queue->current_song = info;
		gdam_producer_queue_set_filename(queue, 
				gdam_song_info_get_filename(info));
	}

	if (g_list_length(songinfo_list) == 2){
		queue_refresh_queued_song(queue);
	}
}

static void print_entry(gpointer data, gpointer count_addr)
{
	GdamSongInfo* song_info = (GdamSongInfo*)data;
	int* n = count_addr; 
	fprintf(stderr, "%d %s\n", (*n)++, 
			gdam_song_info_get_filename(song_info));
}

void gdam_producer_queue_print_queue(GdamProducerQueue*	queue)
{
	int count = 0;
	gdam_debug("printing %d tracks in queue:",
					g_list_length(queue->songinfo_list));
	g_list_foreach(queue->songinfo_list, print_entry, &count);
}

static void gdam_producer_queue_real_skip_to(GdamProducerQueue* queue,
                                             int      	        song_num)
{
	int new_idx;
	gpointer data;
	new_idx = song_num - 1;
	new_idx = CLAMP(new_idx, 0, g_list_length(queue->songinfo_list)-1);
	data = g_list_nth_data(queue->songinfo_list, new_idx);
	if (!data){
		g_warning("Missing data pointer in skip to.");
		return;
	}
	gdam_debug("skip_to song %d of %d", 
				new_idx,
				g_list_length(queue->songinfo_list));
	queue->current_song = (GdamSongInfo*)data;
	gdam_producer_queue_set_filename(queue, 
			gdam_song_info_get_filename((GdamSongInfo*)data));
}

static void gdam_producer_queue_real_skip_songs(GdamProducerQueue* queue,
                                                int      	   num_songs)
{
	int cur_idx, new_idx;
	gpointer data;
	gdam_debug("producer_queue_real_skip_songs: %d", num_songs);
	if (!queue->current_song){
		gdam_debug("no current song, aborting skip.");
		return;
	}
	cur_idx = g_list_index(queue->songinfo_list, queue->current_song);

	if (cur_idx < 0){
		gdam_debug("no cur_idx, aborting skip.");
		return;
	}
	new_idx = cur_idx + num_songs;
	new_idx = CLAMP(new_idx, 0, g_list_length(queue->songinfo_list)-1);
	data = g_list_nth_data(queue->songinfo_list, new_idx);
	if (!data){
		g_warning("Missing data pointer in seek.");
		return;
	}
	gdam_debug("skip_songs to %d of %d", 
				new_idx,
				g_list_length(queue->songinfo_list));
	queue->current_song = (GdamSongInfo*)data;
	gdam_producer_queue_set_filename(queue, 
			gdam_song_info_get_filename((GdamSongInfo*)data));
}

static void current_source_success_callback(int new_id, gpointer user_data){
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(user_data);
	queue->current_id = new_id;
	gtk_object_unref(GTK_OBJECT(queue));
}

static void current_source_failure_callback(gpointer user_data){
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(user_data);
	queue->current_id = 0;
	gtk_object_unref(GTK_OBJECT(queue));
}

/* Called when an mp3 ends.  Assumes that mp3 was the current one,
 * as no other source should experience an eof event. */
static void mp3_ended_callback(gpointer user_data)
{
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(user_data);
	queue->current_song = queue->queued_song;
	queue->current_id = queue->queued_id;
	queue->eof_sub_id = queue->queued_eof_sub_id;

	if (queue->current_song){
		gdam_debug("queue: mp3 ended, had queued song: %s",
			      gdam_song_info_get_filename(queue->current_song));
		queue_refresh_queued_song(queue);
		  /* just want filenames to update, without disturbing the
		   * queue source. */
		queue->ignore_set_filename = TRUE;
		gdam_producer_queue_set_filename(queue, 
			     gdam_song_info_get_filename(queue->current_song));
		queue->ignore_set_filename = FALSE;
	}
	else if (queue->do_repeat){
		gpointer data;
		data = g_list_nth_data(queue->songinfo_list, 0);
		queue->current_song = (GdamSongInfo*)data;
		gdam_debug("queue: mp3 ended, repeating with song: %s",
			      gdam_song_info_get_filename(queue->current_song));
		gdam_producer_queue_set_filename(queue, 
			     gdam_song_info_get_filename(queue->current_song));
	}
	else{
		gdam_debug("queue: mp3 ended, hadn't queued song");
		gdam_producer_source_ended(GDAM_PRODUCER(queue), 0.0);
	}
}

static void eof_connect_success(int eof_sub_id, gpointer user_data)
{
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(user_data);
	queue->eof_sub_id = eof_sub_id;
	gdam_channel_register_handler(
		GDAM_MODEL(queue)->channel,
		gdam_channel_simple_handler_new(eof_sub_id,
						TRUE,
						mp3_ended_callback,
						queue)
	);
}


static void     gdam_producer_queue_real_play  (GdamProducer*    producer,
					        GdamPlayInfo*    play_info)
{
	GdamProducerQueue* queue = GDAM_PRODUCER_QUEUE(producer);
	GdamChannel* channel;
	GdamModel* model;
	GdamRequest* request;
	GdamRequest *prequest, *qrequest;
	int mp3_source_id, queue_source_id;
	int source_id_array[2];
	GdamRequestBuilder* request_builder;
        gdouble  position = play_info ? play_info->position : 0.0;
	int num_args;
	GdamArg args[4];

	g_return_if_fail(queue != NULL);
	if (!queue->current_song) {
		g_warning("gdam_producer_queue_play: no current song");
		return;
	}
	model = GDAM_MODEL(queue);
	if (model->create_source_pending) {
		g_warning("gdam_producer_queue_play: already requesting");
		return;
	}
	channel = model->channel;
	if (!channel){
		g_warning("queue model missing channel!");
		return;
	}

	g_return_if_fail(queue->filename != NULL);

	args[0].type = GDAM_ARG_TYPE_STRING;
	args[0].name = "filename";
	GDAM_ARG_STRING(args + 0) = queue->filename;

	args[1].type = GDAM_ARG_TYPE_INT64;
	args[1].name = "position";
	GDAM_ARG_INT64(args + 1) = (gint64)(position * 44100.0);

	num_args = 2;
	if (queue->buffer_size > 0) {
		args[num_args].type = GDAM_ARG_TYPE_INT;
		args[num_args].name = "buffer_size";
		GDAM_ARG_INT(args + num_args) = queue->buffer_size;
		num_args++;
	}
	if (queue->fade_envelope > 0) {
		args[num_args].type = GDAM_ARG_TYPE_INT;
		args[num_args].name = "fade_envelope";
		GDAM_ARG_INT(args + num_args) = queue->fade_envelope;
		num_args++;
	}
	request = gdam_request_create_source_v(
				"GdamSeekableMp3",
				num_args,
				args,
				current_source_success_callback,
				current_source_failure_callback,
				queue);

	gdam_debug("Queue: creating producing mp3 with position %f", position);
	 /* Ref once for mp3 create callback, 
	  * once for queued mp3 request,
	  * once for queue create callback */
	gtk_object_ref(GTK_OBJECT(queue));
	gtk_object_ref(GTK_OBJECT(queue));
	gtk_object_ref(GTK_OBJECT(queue));
			
	request_builder = gdam_request_builder_new();
	mp3_source_id = gdam_request_builder_append_create(request_builder,
	                                                     request);
	source_id_array[0] = mp3_source_id;

	if (queue->queued_song != NULL){
		qrequest = gdam_request_create_source(
				"GdamSeekableMp3",
				queued_source_success_callback,
				queued_source_failure_callback,
				queue,
				3,
				"filename",
				GDAM_ARG_TYPE_STRING,
				gdam_song_info_get_filename(queue->queued_song),
				"position",
				GDAM_ARG_TYPE_INT64,
				(gint64)0,
				"buffer_size",
				GDAM_ARG_TYPE_INT,
				(gint)queue->buffer_size);

		gdam_debug("Queue: creating queued mp3 with position 0");
				
		source_id_array[1] = gdam_request_builder_append_create(
								request_builder,
								qrequest);
		gdam_request_builder_append_generic(
			request_builder,
			gdam_request_subscribe_source_ended(
				source_id_array[1],
				queue_eof_connect_success,
				NULL,
				queue)
		);
	}
	else{
		source_id_array[1] = 0;
	}

	model->create_source_pending = 1;
	prequest = gdam_request_create_source(
					"GdamQueue",
					gdam_model_got_subsource,
					gdam_model_didnt_get_subsource,
					GDAM_MODEL(queue),
					1,
					"queue",
					GDAM_ARG_TYPE_INT_ARRAY,
					2,
					source_id_array);
	queue_source_id = gdam_request_builder_append_create(request_builder, 
							     prequest);
	gdam_request_builder_append_generic(
		request_builder,
		gdam_request_subscribe_source_ended(
			mp3_source_id,
			eof_connect_success,
			NULL,
			queue)
	);
	gdam_model_new_child_request(model->parent,
	                             model,
				     play_info,
				     queue_source_id,
				     request_builder);
}

static void gdam_producer_queue_destroy(GtkObject* object)
{
	/* TODO: free song infos? */
	(*parent_class->destroy)(object);
}

void gdam_producer_queue_skip_songs(GdamProducerQueue* queue,
				    int 		      num)
{
	gdam_debug("producer_queue_skip_songs");
	gdam_producer_queue_real_skip_songs(queue, num);
}

void gdam_producer_queue_skip_to(GdamProducerQueue* queue,
				 int 		    num)
{
	gdam_debug("producer_queue_skip_to");
	gdam_producer_queue_real_skip_to(queue, num);
}


void gdam_producer_queue_add_filename(GdamProducerQueue* queue,
				      const char*        filename)
{
	gdam_debug("producer_queue_add_filename");
	gdam_producer_queue_real_add_filename(queue, filename);
}

/* This starts the current song playing.  It also updates the queued song.
 * the current_song pointer should have been updated before this is called.
 * TODO: this shouldn't be the main method for switching songs. */
void gdam_producer_queue_set_filename(GdamProducerQueue* queue,
				      const char* 	 filename)
{
	GtkObject* this_object = GTK_OBJECT(queue);

	gdam_debug("producer_queue_set_filename");

	if (gtk_signal_n_emissions(this_object, 
				   producer_queue_signals[SET_FILENAME]))
		return;
	gtk_signal_emit(this_object,
			producer_queue_signals[SET_FILENAME],
			filename);
}

void gdam_producer_queue_set_queue_list
				(GdamProducerQueue* queue,
				 GList*		    songinfo_list)
{
	GtkObject* this_object = GTK_OBJECT(queue);

	gdam_debug("producer_queue_set_queue_list");

	if (gtk_signal_n_emissions(this_object, 
				   producer_queue_signals[SET_QUEUE_LIST]))
		return;
	gtk_signal_emit(this_object,
			producer_queue_signals[SET_QUEUE_LIST],
			songinfo_list);
}

static void gdam_producer_queue_real_set_queue_list(GdamProducerQueue* queue,
						    gpointer 	       data)
{
	GList*      songinfo_list = (GList*)data;
	gboolean had_current = queue->current_song ? TRUE : FALSE;
	gdam_debug("producer_queue_real_set_queue_list");
	queue->songinfo_list = g_list_copy(songinfo_list);
	if (g_list_length(songinfo_list) <= 0){
		queue->current_song = NULL;
		queue->queued_song  = NULL;
		gdam_producer_stop(GDAM_PRODUCER(queue));
	}
	else if (!had_current){
		gpointer data;
		data = g_list_nth_data(queue->songinfo_list, 0);
		queue->current_song = (GdamSongInfo*)data;
		gdam_producer_queue_set_filename(queue,
			      gdam_song_info_get_filename(queue->current_song));
	}
	else {
		int new_idx;
		new_idx = g_list_index(songinfo_list, queue->current_song);
		/* TODO.... maybe shouldn't skip to first song? */
		if (new_idx < 0){
			gpointer data;
			data = g_list_nth_data(queue->songinfo_list, 0);
			queue->current_song = (GdamSongInfo*)data;
			gdam_producer_queue_set_filename(queue,
				      gdam_song_info_get_filename(
					      		queue->current_song));
		}
		else
			queue_refresh_queued_song(queue);
	}
}

static void gdam_producer_queue_real_set_filename(GdamProducerQueue* queue,
                                                  const char*      filename)
{
	GdamSongDatabase* database = gdam_song_database_system();
	GdamSongInfo* info = NULL;
	GdamBeatInfo beat_info;
	gboolean was_playing = FALSE;
	gdam_debug("producer_queue_real_set_filename");

	queue_refresh_queued_song(queue);

	queue->filename = g_strdup(filename);

	/* HACK to alert listeners of changed filename when switching to
	 * queued_source, without causing song to replay. */
	if (queue->ignore_set_filename)
		return;

	        /* TODO: might this permit multiple requests? */
	if (GDAM_MODEL(queue)->source_id != 0){
		was_playing = TRUE;
		gdam_producer_stop(GDAM_PRODUCER(queue));
	}
	if (database)
		info = gdam_song_database_find(database, filename);
	gdam_beat_info_set_first_beat(&beat_info, 0);
	gdam_beat_info_set_bpm(&beat_info, 0);
	if (info) {
		gdam_producer_update_length(GDAM_PRODUCER(queue), 
					    gdam_song_info_get_length(info));
		beat_info = *gdam_song_info_get_beat_info(info);
	}
	gdam_producer_set_position(GDAM_PRODUCER(queue), 0);
	gdam_model_set_beat_info(GDAM_MODEL(queue), &beat_info);
	if (was_playing)
		gdam_producer_play_now(GDAM_PRODUCER(queue));
}

GdamProducerQueue* gdam_producer_queue_new(GdamChannel*	channel)
{
	GdamProducerQueue* rv;
	rv = GDAM_PRODUCER_QUEUE(gtk_type_new(GDAM_TYPE_PRODUCER_QUEUE));
	((GdamModel*)rv)->channel = channel;
	return rv;
}

void gdam_producer_queue_print_range(GdamProducerQueue* queue,
				     guint lower, guint upper, guint dest)
{
} 


void gdam_producer_queue_search_by_string
				    (GdamProducerQueue* queue,
				     const char*	search_string,
				     int		direction)
{
	GList* cur;
	GList* orig;
	if (g_list_length(queue->songinfo_list) == 0)
		return;
	cur = orig = queue->songinfo_list;
	if (direction > 0)
		cur = cur->next ? cur->next : queue->songinfo_list;
	else 
		cur = cur->prev ? cur->prev : g_list_last(queue->songinfo_list);

	while (cur){
		if (cur == orig)
			cur = NULL;
		else {
			GdamSongInfo* song = (GdamSongInfo*)(cur->data);
			if (strstr( gdam_song_info_get_filename(song), 
				    search_string)){
				queue->current_song = (GdamSongInfo*)cur->data;
				gdam_producer_queue_set_filename( queue,
					     gdam_song_info_get_filename(
							  queue->current_song));
				cur = NULL;
			}
			else if (direction > 0)
				cur = cur->next ? 
				      cur->next : 
				      queue->songinfo_list;
			else
				cur = cur->prev ? 
				      cur->prev : 
				      g_list_last(queue->songinfo_list);
		}
	}
}

void gdam_queue_arg_set_queue_list(GtkObject* object, GdamArg* arg)
{
	GdamProducerQueue* queue;
	char* nl;
	char* cur_file;
	char* filenames = g_strdup(GDAM_ARG_STRING(arg));
	gboolean keep_going = TRUE;
	g_message("set_queue_list called...");
	queue = GDAM_PRODUCER_QUEUE(object);
	cur_file = filenames;
	g_list_free(queue->songinfo_list);
	queue->songinfo_list = NULL;
	while (keep_going && *cur_file != '\0'){
		nl = strchr(cur_file, '\n');
		if (nl)
			*nl = '\0';
		else
			keep_going = FALSE;
		gdam_debug("arg set queue list: adding %s", cur_file);
		gdam_producer_queue_add_filename(queue, cur_file);
		cur_file = nl+1;
	}
}

static void gdam_queue_arg_get_queue_list(GtkObject* object, GdamArg* arg)
{
	GdamProducerQueue* queue;
	char str[16383] = "";
	GList* l;
	queue = GDAM_PRODUCER_QUEUE(object);
	for (l=queue->songinfo_list;l!=NULL;l=l->next){
		strcat(str,
		       gdam_song_info_get_filename((GdamSongInfo*)(l->data)));
		strcat(str, "\n");
	}
	GDAM_ARG_STRING(arg) = strdup(str);
}

static void gdam_producer_queue_class_init(GdamProducerQueueClass* klass)
{
	GdamProducerClass* producer_class = GDAM_PRODUCER_CLASS(klass);
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	klass->skip_songs       = gdam_producer_queue_real_skip_songs;
	klass->skip_to          = gdam_producer_queue_real_skip_to;
	klass->add_filename = gdam_producer_queue_real_add_filename;
	klass->set_filename = gdam_producer_queue_real_set_filename;
	klass->set_queue_list = gdam_producer_queue_real_set_queue_list;
	producer_class->play = gdam_producer_queue_real_play;
	object_class->destroy   = gdam_producer_queue_destroy;
	parent_class = gtk_type_class(GDAM_TYPE_PRODUCER);
	gdam_arg_register(object_class,
			  "queue_list",
			  GDAM_ARG_TYPE_STRING,
			  gdam_queue_arg_set_queue_list,
			  gdam_queue_arg_get_queue_list,
			  GDAM_ARG_VITAL);
	producer_queue_signals[SET_QUEUE_LIST] =
		gtk_signal_new(
			"set_queue_list",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamProducerQueueClass, 
					  set_queue_list),
			gtk_marshal_NONE__POINTER,
			GTK_TYPE_NONE,
			1,
			GTK_TYPE_POINTER
		);
	producer_queue_signals[SET_FILENAME] =
		gtk_signal_new(
			"set_filename",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamProducerQueueClass, set_filename),
			gtk_marshal_NONE__STRING,
			GTK_TYPE_NONE,
			1,
			GTK_TYPE_STRING
		);
	gtk_object_class_add_signals(object_class,
	                             producer_queue_signals,
				     LAST_SIGNAL);
}
