/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdamequalizer.h"
#include <gdam/gdamsample.h>
#include <gtk/gtksignal.h>
#include <math.h>
#include "../gdammodel.h"
#include "../gdamfilter.h"
#include "../gdamutil.h"
static void gdam_equalizer_class_init(GdamEqualizerClass* klass);

/* Filter class function */
static int gdam_equalizer_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder);

static void gdam_equalizer_init(GdamEqualizer* equalizer)
{
	equalizer->frequency = 440; 
	equalizer->mode = "peakingeq";
	equalizer->db_gain = equalizer->bandwidth = 1.0;
	equalizer->effect = 0.0;
} 

GtkType gdam_equalizer_get_type()
{
	static GtkType equalizer_type = 0;
	if (!equalizer_type) {
		static const GtkTypeInfo equalizer_info = {
			"GdamEqualizer",
			sizeof(GdamEqualizer),
			sizeof(GdamEqualizerClass),
			(GtkClassInitFunc) gdam_equalizer_class_init,
			(GtkObjectInitFunc) gdam_equalizer_init,
			/* reserved 1 */ NULL,
			/* reserved 2 */ NULL,
			(GtkClassInitFunc) NULL
		};
		equalizer_type = gtk_type_unique(GDAM_TYPE_FILTER, &equalizer_info);
	}
	return equalizer_type;
}

#define NUM_ARGS	5
static inline void build_args_array(GdamEqualizer* equalizer,
                                    GdamArg*       args)
{
	args[0].name = "frequency";
	args[0].type = GDAM_ARG_TYPE_DOUBLE;
	GDAM_ARG_DOUBLE(args+0) = equalizer->frequency;
	args[1].name = "db_gain";
	args[1].type = GDAM_ARG_TYPE_DOUBLE;
	GDAM_ARG_DOUBLE(args+1) = equalizer->db_gain;
	args[2].name = "bandwidth";
	args[2].type = GDAM_ARG_TYPE_DOUBLE;
	GDAM_ARG_DOUBLE(args+2) = equalizer->bandwidth;
	args[3].name = "mode";
	args[3].type = GDAM_ARG_TYPE_STRING;
	GDAM_ARG_STRING(args+3) = (char*) equalizer->mode;
	args[4].name = "effect";
	args[4].type = GDAM_ARG_TYPE_DOUBLE;
	GDAM_ARG_DOUBLE(args+4) = equalizer->effect;
}



static void maybe_dispense_request(GdamEqualizer* equalizer)
{
	GdamRequest* request;
	GdamArg args[5];
	GdamModel* model = GDAM_MODEL(equalizer);
	int source_id = model->source_id;
	if (source_id == 0){
		gdam_debug("equalizer maybe_dispense_request aborting");
		return;
	}

	build_args_array(equalizer, args);

	request = gdam_request_set_args(
			source_id,
			NUM_ARGS,
			args,
			NULL, NULL, NULL);
	gdam_channel_append_request(model->channel, request);
}

GdamEqualizer* 	gdam_equalizer_new		()
{
	GdamEqualizer* equalizer;
	equalizer = GDAM_EQUALIZER(gtk_type_new(GDAM_TYPE_EQUALIZER));
	return equalizer;
}

static int gdam_equalizer_append_start_request(GdamFilter* filter,
				          int new_source_id,
				          GdamRequestBuilder* builder)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(filter);
	GdamRequest* request;
	GdamArg args[NUM_ARGS + 1];
	gtk_object_ref(GTK_OBJECT(equalizer));
	GDAM_MODEL(equalizer)->create_source_pending = 1;
	build_args_array(equalizer, args);
	args[NUM_ARGS].name = "filtered";
	args[NUM_ARGS].type = GDAM_ARG_TYPE_INT;
	GDAM_ARG_INT(args+NUM_ARGS) = new_source_id;
	request = gdam_request_create_source_v
					("GdamEqualizer",
					 NUM_ARGS + 1,
					 args,
					 gdam_model_got_subsource,
					 gdam_model_didnt_get_subsource,
					 equalizer);
	return gdam_request_builder_append_create(builder, request);
}

static void gdam_equalizer_arg_set_frequency(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	equalizer->frequency = GDAM_ARG_DOUBLE(arg);
	maybe_dispense_request(equalizer);
}
static void gdam_equalizer_arg_get_frequency(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	GDAM_ARG_DOUBLE(arg) = equalizer->frequency;
}

static const char* get_mode_name_from_name(const char* name)
{
	static const char* names[] = {
		"lowpass",
		"highpass",
		"bandpass",
		"peakingeq",
		"lowshelf",
		"highshelf",
		"notch",
		NULL
	};
	int i;
	for (i = 0; names[i] != NULL; i++) {
		if (strcmp(names[i], name) == 0)
			return names[i];
	}
	return NULL;
}

static void gdam_equalizer_arg_set_mode(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	const char* mode = get_mode_name_from_name(GDAM_ARG_STRING(arg));
	g_return_if_fail(mode != NULL);
	equalizer->mode = mode;
	maybe_dispense_request(equalizer);
}
static void gdam_equalizer_arg_get_mode(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	GDAM_ARG_STRING(arg) = g_strdup(equalizer->mode);
}

static void gdam_equalizer_arg_set_db_gain(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	equalizer->db_gain = GDAM_ARG_DOUBLE(arg);
	maybe_dispense_request(equalizer);
}
static void gdam_equalizer_arg_get_db_gain(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	GDAM_ARG_DOUBLE(arg) = equalizer->db_gain;
}

static void gdam_equalizer_arg_set_bandwidth(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	equalizer->bandwidth = GDAM_ARG_DOUBLE(arg);
	maybe_dispense_request(equalizer);
}
static void gdam_equalizer_arg_get_bandwidth(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	GDAM_ARG_DOUBLE(arg) = equalizer->bandwidth;
}
static void gdam_equalizer_arg_set_effect(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	equalizer->effect = GDAM_ARG_DOUBLE(arg);
	maybe_dispense_request(equalizer);
}
static void gdam_equalizer_arg_get_effect(GtkObject* object, GdamArg* arg)
{
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	GDAM_ARG_DOUBLE(arg) = equalizer->effect;
}

static void gdam_equalizer_arg_set_octave(GtkObject* object, GdamArg* arg)
{
	gdouble octave = GDAM_ARG_DOUBLE(arg);
	gdouble frequency = pow(2.0, octave) * 440.0;
	GdamEqualizer* equalizer = GDAM_EQUALIZER(object);
	equalizer->frequency = frequency;
	maybe_dispense_request(equalizer);
}

static void gdam_equalizer_class_init(GdamEqualizerClass* klass)
{
	GdamFilterClass* filter_class;
	GdamModelClass* model_class;
	GtkObjectClass* object_class;
	filter_class = GDAM_FILTER_CLASS(klass);
	model_class = GDAM_MODEL_CLASS(klass);
	object_class = GTK_OBJECT_CLASS(klass);
	filter_class->append_start_request
		= gdam_equalizer_append_start_request;

	gdam_arg_register(object_class,
	                  "db_gain",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_equalizer_arg_set_db_gain,
			  gdam_equalizer_arg_get_db_gain,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "frequency",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_equalizer_arg_set_frequency,
			  gdam_equalizer_arg_get_frequency,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "octave",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_equalizer_arg_set_octave,
			  NULL,
			  0);
	gdam_arg_register(object_class,
	                  "bandwidth",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_equalizer_arg_set_bandwidth,
			  gdam_equalizer_arg_get_bandwidth,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "effect",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_equalizer_arg_set_effect,
			  gdam_equalizer_arg_get_effect,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "mode",
			  GDAM_ARG_TYPE_STRING,
			  gdam_equalizer_arg_set_mode,
			  gdam_equalizer_arg_get_mode,
			  GDAM_ARG_VITAL);
}

