/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamflange.h"
static void gdam_flange_factors_changed(GdamFlange* flange);

static void gdam_flange_init(GdamFlange* flange)
{
	flange->hertz = 1.0;
	flange->buffer_id = 1;
	GDAM_FIR_FILTER(flange)->grid_unit = 16.0;
}

void gdam_flange_set_min_factor(GdamFlange* flange, double factor)
{
	GdamArg arg;
	arg.name = "min_factor";
	arg.type = GDAM_ARG_TYPE_DOUBLE;
	GDAM_ARG_DOUBLE(&arg) = factor;
	gdam_arg_set_object_arg(GTK_OBJECT(flange), &arg);
}

void gdam_flange_set_max_factor(GdamFlange* flange, double factor)
{
	GdamArg arg;
	arg.name = "max_factor";
	arg.type = GDAM_ARG_TYPE_DOUBLE;
	GDAM_ARG_DOUBLE(&arg) = factor;
	gdam_arg_set_object_arg(GTK_OBJECT(flange), &arg);
}

static void gdam_flange_arg_set_min_factor(GtkObject* object, GdamArg* arg)
{
	GdamFlange* flange = GDAM_FLANGE(object);
	gdouble center_factor = (flange->min_factor + flange->max_factor) / 2.0;
	gdouble max_factor;
	flange->min_factor = GDAM_ARG_DOUBLE(arg);
	if (flange->min_factor > center_factor)
		flange->min_factor = center_factor;
	max_factor = center_factor + (center_factor - flange->min_factor);
	if (max_factor != flange->max_factor)
		gdam_flange_set_max_factor(flange, max_factor);
	//flange->max_factor = max_factor;
	gdam_flange_factors_changed(flange);
}
static void gdam_flange_arg_get_min_factor(GtkObject* object, GdamArg* arg)
{
	GdamFlange* flange = GDAM_FLANGE(object);
	GDAM_ARG_DOUBLE(arg) = flange->min_factor;
}

static void gdam_flange_arg_set_mid_factor(GtkObject* object, GdamArg* arg)
{
	GdamFlange* flange = GDAM_FLANGE(object);
	gdouble old_center_factor;
	gdouble delta;
	gdouble min_factor, max_factor;

	old_center_factor = (flange->min_factor + flange->max_factor) / 2.0;
	delta = GDAM_ARG_DOUBLE(arg) - old_center_factor;
	min_factor = flange->min_factor + delta;
	max_factor = flange->max_factor + delta;
	flange->min_factor = min_factor;
	flange->max_factor = max_factor;
	gdam_flange_set_min_factor(flange, min_factor);
	gdam_flange_set_max_factor(flange, max_factor);
	gdam_flange_factors_changed(flange);
}
static void gdam_flange_arg_get_mid_factor(GtkObject* object, GdamArg* arg)
{
	GdamFlange* flange = GDAM_FLANGE(object);
	GDAM_ARG_DOUBLE(arg) = (flange->min_factor + flange->max_factor) / 2.0;
}

static void gdam_flange_arg_set_max_factor(GtkObject* object, GdamArg* arg)
{
	GdamFlange* flange = GDAM_FLANGE(object);
	gdouble center_factor = (flange->min_factor + flange->max_factor) / 2.0;
	gdouble min_factor;
	flange->max_factor = GDAM_ARG_DOUBLE(arg);
	if (flange->max_factor < center_factor)
		flange->max_factor = center_factor;
	min_factor = center_factor - (flange->max_factor - center_factor);
	if (min_factor != flange->min_factor)
		gdam_flange_set_min_factor(flange, min_factor);
	//flange->min_factor = min_factor;
	gdam_flange_factors_changed(flange);
}
static void gdam_flange_arg_get_max_factor(GtkObject* object, GdamArg* arg)
{
	GdamFlange* flange = GDAM_FLANGE(object);
	GDAM_ARG_DOUBLE(arg) = flange->max_factor;
}

static int
make_wavetable_request(GdamFlange* flange, GdamRequestBuilder* builder)
{
	GdamRequest* request;
	request = gdam_request_create_source
				("GdamWavetable",
				 NULL,
				 NULL,
				 NULL,
				 2,
				 "buffer_id",
			         GDAM_ARG_TYPE_INT,
				 flange->buffer_id,
				 "hertz",
				 GDAM_ARG_TYPE_DOUBLE,
				 flange->hertz);
	return gdam_request_builder_append_create(builder, request);
}

static int gdam_flange_append_start_request(GdamFilter* filter,
				            int new_source_id,
				            GdamRequestBuilder* builder)
{
	GdamEchoBase* echo_base = GDAM_ECHO_BASE(filter);
	GdamRequest* request;
	int wavetable_id;
	gdouble min_max[2];
	GdamFlange* flange = GDAM_FLANGE(filter);
	gtk_object_ref(GTK_OBJECT(filter));
	GDAM_MODEL(filter)->create_source_pending = 1;
	min_max[0] = flange->min_factor;
	min_max[1] = flange->max_factor;
	wavetable_id = make_wavetable_request(flange, builder);
	request = gdam_request_create_source("GdamFlange",
					     gdam_model_got_subsource,
					     gdam_model_didnt_get_subsource,
					     flange,
					     4,
					     "filtered",
					     GDAM_ARG_TYPE_INT,
					     new_source_id,
					     "modulator",
					     GDAM_ARG_TYPE_INT,
					     wavetable_id,
					     "echoes",
					     GDAM_ARG_TYPE_INT_ARRAY,
					     echo_base->num_echoes * 2,
					     echo_base->echoes,
					     "min_max",
					     GDAM_ARG_TYPE_DOUBLE_ARRAY,
					     2,
					     min_max);
	return gdam_request_builder_append_create(builder, request);
}

static void reset_wavetable(GdamFlange* flange)
{
	GdamModel* model = (GdamModel*) flange;
	if (model->source_id != 0) {
		GdamRequestBuilder* builder;
		GdamArg arg;
		GdamRequest* package;
		int wavetable_id;
		builder = gdam_request_builder_new();
		wavetable_id = make_wavetable_request(flange, builder);
		arg.name = "modulator";
		arg.type = GDAM_ARG_TYPE_INT;
		GDAM_ARG_INT(&arg) = wavetable_id;
		gdam_request_builder_append_generic
			(builder,
			 gdam_request_set_args
				(model->source_id,
				 1,
				 &arg,
				 NULL, NULL, NULL));
		package = gdam_request_builder_package(builder);
		gdam_channel_append_request(model->channel, package);
		gdam_request_builder_destroy(builder);
	} else if (model->create_source_pending) {
		g_warning("bug: cannot set wavetable right now...");
	}
}
static void gdam_flange_arg_set_buffer_id(GtkObject* object, GdamArg* arg)
{
	GdamFlange* flange = GDAM_FLANGE(object);
	int buffer_id = GDAM_ARG_INT(arg);
	g_message("got set_buffer_id %d", buffer_id);
	if (buffer_id != flange->buffer_id) {
		flange->buffer_id = buffer_id;
		reset_wavetable(flange);
	}
}
static void gdam_flange_arg_get_buffer_id(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_FLANGE(object)->buffer_id;
}
static void gdam_flange_arg_set_hertz(GtkObject* object, GdamArg* arg)
{
	GdamFlange* flange = GDAM_FLANGE(object);
	gdouble hertz = GDAM_ARG_DOUBLE(arg);
	if (hertz != flange->hertz) {
		flange->hertz = hertz;
		reset_wavetable(flange);
	}
}
static void gdam_flange_arg_get_hertz(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_DOUBLE(arg) = GDAM_FLANGE(object)->hertz;
}


static void gdam_flange_class_init(GdamFlangeClass* flange_class)
{
	GtkObjectClass* object_class;
	object_class = GTK_OBJECT_CLASS(flange_class);
	gdam_arg_register(object_class,
	                  "min_factor",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_flange_arg_set_min_factor,
			  gdam_flange_arg_get_min_factor,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "mid_factor",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_flange_arg_set_mid_factor,
			  gdam_flange_arg_get_mid_factor,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "max_factor",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_flange_arg_set_max_factor,
			  gdam_flange_arg_get_max_factor,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "buffer_id",
			  GDAM_ARG_TYPE_INT,
			  gdam_flange_arg_set_buffer_id,
			  gdam_flange_arg_get_buffer_id,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "hertz",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_flange_arg_set_hertz,
			  gdam_flange_arg_get_hertz,
			  GDAM_ARG_VITAL);
	GDAM_FILTER_CLASS(flange_class)->append_start_request
		= gdam_flange_append_start_request;
}

GtkType gdam_flange_get_type()
{
	static GtkType flange_type = 0;
	if (!flange_type) {
		static const GtkTypeInfo flange_info = {
			"GdamFlange",
			sizeof(GdamFlange),
			sizeof(GdamFlangeClass),
			(GtkClassInitFunc) gdam_flange_class_init,
			(GtkObjectInitFunc) gdam_flange_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_FIR_FILTER;
		flange_type = gtk_type_unique(parent,
		                                 &flange_info);
	}
	return flange_type;
}

static void gdam_flange_factors_changed(GdamFlange* flange)
{
	GdamModel* model = GDAM_MODEL(flange);
	guint source_id = model->source_id;
	if (source_id != 0) {
		GdamArg args[2];
		gdouble min_max[2];
		args[0].name = "min_max";
		args[0].type = GDAM_ARG_TYPE_DOUBLE_ARRAY;
		GDAM_ARG_DOUBLE_ARRAY(args+0).array = min_max;
		GDAM_ARG_DOUBLE_ARRAY(args+0).length = 2;
		min_max[0] = flange->min_factor;
		min_max[1] = flange->max_factor;
		gdam_channel_append_request(
			model->channel,
			gdam_request_set_args
				(source_id,
				 1,
				 args,
				 NULL,
				 NULL,
				 NULL)
		);
	}
}
