/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include "gdamproducerwavetable.h"
#include "../gdambuffermanager.h"
#include <gtk/gtksignal.h>
#include <gdam/gdamplugin.h>
#include "../gdamutil.h"

enum {
	SET_BUFFER_ID,
	RATE_CHANGED,
	LAST_SIGNAL
};

static guint 	producer_wavetable_signals[LAST_SIGNAL];

static void     gdam_producer_wavetable_play       (GdamProducer*   producer,
					            GdamPlayInfo*   play_info);
static void 
gdam_producer_wavetable_real_set_buffer_id  (GdamProducerWavetable* wavetable,
                                             int      		new_buffer_id);

static void
gdam_producer_wavetable_detach_from_buffer(GdamModel*        model,
                                           guint             buffer_id);
static void 
gdam_producer_wavetable_init	(GdamProducerWavetable* producer_wavetable)
{
	producer_wavetable->hertz = 440.0;
	producer_wavetable->buffer_id = -1;
	producer_wavetable->making_play_request = 0;
}

static void 
gdam_producer_wavetable_class_init	(GdamProducerWavetableClass* klass)
{
	GtkObjectClass* object_class = GTK_OBJECT_CLASS(klass);
	klass->set_buffer_id = gdam_producer_wavetable_real_set_buffer_id;
	GDAM_PRODUCER_CLASS(klass)->play = gdam_producer_wavetable_play;
	GDAM_MODEL_CLASS(klass)->detach_from_buffer
		= gdam_producer_wavetable_detach_from_buffer;
	producer_wavetable_signals[SET_BUFFER_ID] =
		gtk_signal_new(
			"set_buffer_id",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamProducerWavetableClass, 
					  set_buffer_id),
			gtk_marshal_NONE__INT,
			GTK_TYPE_NONE,
			1,
			GTK_TYPE_INT
		);
	producer_wavetable_signals[RATE_CHANGED] =
		gtk_signal_new(
			"rate_changed",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(GdamProducerWavetableClass, 
					  rate_changed),
			gtk_marshal_NONE__NONE,
			GTK_TYPE_NONE,
			0
		);
	gtk_object_class_add_signals(object_class,
	                             producer_wavetable_signals,
				     LAST_SIGNAL);
}

GtkType gdam_producer_wavetable_get_type()
{
	static GtkType producer_wavetable_type = 0;
	if (!producer_wavetable_type) {
		static const GtkTypeInfo producer_wavetable_info = {
			"GdamProducerWavetable",
			sizeof(GdamProducerWavetable),
			sizeof(GdamProducerWavetableClass),
			(GtkClassInitFunc) gdam_producer_wavetable_class_init,
			(GtkObjectInitFunc) gdam_producer_wavetable_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = gdam_plugin_get_type_by_name("component",
							      "GdamProducer");
		producer_wavetable_type = gtk_type_unique(parent,
		                                 &producer_wavetable_info);
	}
	return producer_wavetable_type;
}

static void     gdam_producer_wavetable_play        (GdamProducer* producer,
						     GdamPlayInfo* play_info)
{
	GdamProducerWavetable* wavetable = GDAM_PRODUCER_WAVETABLE(producer);
	GdamChannel* channel;
	GdamModel* component;
	GdamRequestBuilder* request_builder;
	GdamRequest* request;
	gdouble position;
	int local_source_id;

	g_return_if_fail(wavetable != NULL);

	if (wavetable->buffer_id < 0) {
		g_warning("gdam_producer_waveform_play: no buffer_id");
		return;
	}
	component = GDAM_MODEL(wavetable);
	channel = component->channel;
	if (!channel){
		g_warning("wavetable component missing channel!");
		return;
	}

	position = play_info->position;

	gdam_debug("Creating producing wavetable with position %f", position);
	gtk_object_ref(GTK_OBJECT(wavetable));
	component->create_source_pending = 1;
	request = gdam_request_create_source(
				"GdamWavetable",
				gdam_model_got_subsource,
				gdam_model_didnt_get_subsource,
				wavetable,
				2,
				"buffer_id",   
				GDAM_ARG_TYPE_INT,
				wavetable->buffer_id,
				"position",
				GDAM_ARG_TYPE_INT64,
				(gint64)(position * 44100.0));

	request_builder = gdam_request_builder_new();
	local_source_id = gdam_request_builder_append_create(request_builder,
	                                                     request);

	gdam_model_new_child_request(component->parent,
	                                 component,
					 play_info,
				         local_source_id,
				         request_builder);
}

static void
gdam_producer_wavetable_detach_from_buffer(GdamModel*        model,
                                           guint             buffer_id)
{
	GdamProducerWavetable* wavetable;
	wavetable = GDAM_PRODUCER_WAVETABLE(model);
	if (wavetable->buffer_id != 0)
		gdam_producer_wavetable_set_buffer_id(wavetable, 0);
}

static void 
gdam_producer_wavetable_real_set_buffer_id (GdamProducerWavetable* wavetable,
                                            int      		  new_buffer_id)
{
	/*GdamBeatInfo beat_info;*/
	GdamBufferManagerInfo* info;
	GdamModel* model = GDAM_MODEL(wavetable);
	gdam_producer_stop(GDAM_PRODUCER(wavetable));

	gdam_buffer_manager_ref(new_buffer_id, model);
	if (wavetable->buffer_id > 0)
		gdam_buffer_manager_unref(wavetable->buffer_id, model);

	wavetable->buffer_id = new_buffer_id;
	info = gdam_buffer_manager_info_by_id(new_buffer_id);

	if (info == NULL) {
		wavetable->buffer_length = 0;
		return;
	}
	wavetable->buffer_length = info->length;
	if (info->beat_info) {
		gdam_model_set_beat_info(GDAM_MODEL(wavetable),
					     info->beat_info);
	} else {
		gdam_model_set_beat_info(GDAM_MODEL(wavetable),
					     gdam_beat_info_new());
	}

	/*
	gtk_signal_emit_by_name(GTK_OBJECT(wavetable),
				"update_length",
				 info->length);
	beat_info.first_beat = info->first_beat;
	beat_info.bpm = info->bpm;
	gdam_model_set_beat_info(GDAM_MODEL(wavetable), &beat_info);
	 */
}

void gdam_producer_wavetable_set_buffer_id(GdamProducerWavetable* wavetable,
                                           int                    buffer_id)
{
	GtkObject* this_object = GTK_OBJECT(wavetable);

	if (gtk_signal_n_emissions(this_object,
	                           producer_wavetable_signals[SET_BUFFER_ID]))
		return;

	gtk_signal_emit(this_object,
	   	    	producer_wavetable_signals[SET_BUFFER_ID],
			buffer_id);
}

void gdam_producer_wavetable_set_frequency(GdamProducerWavetable* wavetable,
                                           gdouble                hertz)
{
	wavetable->hertz = hertz;
	wavetable->rate = (gdouble) wavetable->buffer_length
	                          / (44100.0 / wavetable->hertz);
	gdam_producer_wavetable_frequency_changed(wavetable);
}
void gdam_producer_wavetable_set_rate     (GdamProducerWavetable* wavetable,
                                           gdouble                rate)
{
	wavetable->rate = rate;
	wavetable->hertz = (gdouble) 44100.0 / (wavetable->buffer_length/rate);
	gdam_producer_wavetable_frequency_changed(wavetable);
}


void gdam_producer_wavetable_frequency_changed(GdamProducerWavetable* wavetable)
{
	GtkObject* object = GTK_OBJECT(wavetable);
	GdamArg arg;
	GdamRequest* request;
	GdamModel* component = GDAM_MODEL(wavetable);
	guint signal = producer_wavetable_signals[RATE_CHANGED];
	if (gtk_signal_n_emissions(object, signal))
		return;
	if (!component->source_id)
		return;

	arg.name = "hertz";
	arg.type = GDAM_ARG_TYPE_DOUBLE;
	GDAM_ARG_DOUBLE(&arg) = wavetable->hertz;
	request = gdam_request_set_args(
			component->source_id,
			1,
			&arg,
			gdam_do_nothing,
			gdam_do_nothing,
			NULL);

	gdam_channel_append_request(component->channel, request);
	gtk_signal_emit(object, signal);
}




