/* CERES: A combination of FFT and granular synthesis */

#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <audiofile.h>
#include <audio.h>
#include <sys/fpu.h>


/* Flags to indicate whether the fft is forward or reverse. */
#define INVERSE 		0
#define FORWARD 		1

/* Max allowed fft window. */
#define MAX_FFT_SIZE 		4096

#define MAXNOD 800


#define fvec(name, size)		(name = g_new0(gdouble, size))
#define ffvec(name, size)		(name = g_new0(gdouble, size))

static gdouble* new_double_array(int num);

#define LSIZE 512

extern double lpamp(double, double, double *, int);
extern double lpa(double *, int, double *, int, int);


#define PI 3.141592654
#define TWOPI 6.2831853

/* FFT ROUTINES */

/* If forward is true, rfft replaces 2*N real data points in x with
   N complex values representing the positive frequency half of their
   Fourier spectrum, with x[1] replaced with the real part of the Nyquist
   frequency value.  If forward is false, rfft expects x to contain a
   positive frequency spectrum arranged as before, and replaces it with
   2*N real values.  N MUST be a power of 2. */

static void rfft( double* x, int N, gboolean forward)
{
	double    c1, c2, h1r, h1i, h2r, h2i, wr, wi, wpr, wpi, temp, theta;
	double    xr, xi;
	int       i, i1, i2, i3, i4, N2p1;

	theta = PI / N;
	wr = 1.0;
	wi = 0.0;
	c1 = 0.5;
	if (forward) {
		c2 = -0.5;
		cfft(x, N, forward);
		xr = x[0];
		xi = x[1];
	} else {
		c2 = 0.5;
		theta = -theta;
		xr = x[1];
		xi = 0.;
		x[1] = 0.;
	}
	wpr = -2. * powf(sin(0.5 * theta), 2.);
	wpi = sin(theta);
	N2p1 = (N << 1) + 1;
	for (i = 0; i <= N >> 1; i++) {
		i1 = i << 1;
		i2 = i1 + 1;
		i3 = N2p1 - i2;
		i4 = i3 + 1;
		if (i == 0) {
			h1r = c1 * (x[i1] + xr);
			h1i = c1 * (x[i2] - xi);
			h2r = -c2 * (x[i2] + xi);
			h2i = c2 * (x[i1] - xr);
			x[i1] = h1r + wr * h2r - wi * h2i;
			x[i2] = h1i + wr * h2i + wi * h2r;
			xr = h1r - wr * h2r + wi * h2i;
			xi = -h1i + wr * h2i + wi * h2r;
		} else {
			h1r = c1 * (x[i1] + x[i3]);
			h1i = c1 * (x[i2] - x[i4]);
			h2r = -c2 * (x[i2] + x[i4]);
			h2i = c2 * (x[i1] - x[i3]);
			x[i1] = h1r + wr * h2r - wi * h2i;
			x[i2] = h1i + wr * h2i + wi * h2r;
			x[i3] = h1r - wr * h2r + wi * h2i;
			x[i4] = -h1i + wr * h2i + wi * h2r;
		}
		wr = (temp = wr) * wpr - wi * wpi + wr;
		wi = wi * wpr + temp * wpi + wi;
	}
	if (forward)
		x[1] = xr;
	else
		cfft(x, N, forward);
}

/*
 * cfft replaces double array x containing NC complex values (2*NC double
 * values alternating real, imagininary, etc.) by its Fourier transform if
 * forward is true, or by its inverse Fourier transform if forward is false,
 * using a recursive Fast Fourier transform method due to Danielson and
 * Lanczos.  NC MUST be a power of 2. 
 */

static void cfft(gdouble* x, int NC, gboolean forward)
{
	gdouble wr, wi, wpr, wpi, theta, scale;
	int     mmax, ND, m, i, j, delta;

	ND = NC << 1;
	bitreverse(x, ND);
	for (mmax = 2; mmax < ND; mmax = delta) {
		delta = mmax << 1;
		theta = TWOPI / (forward ? mmax : -mmax);
		wpr = -2. * powf(sin(0.5 * theta), 2.);
		wpi = sin(theta);
		wr = 1.;
		wi = 0.;
		for (m = 0; m < mmax; m += 2) {
			double rtemp, itemp;
			for (i = m; i < ND; i += delta) {
				j = i + mmax;
				rtemp = wr * x[j] - wi * x[j + 1];
				itemp = wr * x[j + 1] + wi * x[j];
				x[j] = x[i] - rtemp;
				x[j + 1] = x[i + 1] - itemp;
				x[i] += rtemp;
				x[i + 1] += itemp;
			}
			wr = (rtemp = wr) * wpr - wi * wpi + wr;
			wi = wi * wpr + rtemp * wpi + wi;
		}
	}

	/*
	 * scale output 
	 */

	scale = forward ? 1. / ND : 2.;
	for (i = 0; i < ND; i++)
		x[i] *= scale;
}

/*
 * bitreverse places double array x containing N/2 complex values into
 * bit-reversed order 
 */

static void bitreverse(double x[], int N)
{
	double          rtemp, itemp;
	int             i, j, m;

	for (i = j = 0; i < N; i += 2, j += m) {
		if (j > i) {
			rtemp = x[j];
			itemp = x[j + 1];	/* complex exchange */
			x[j] = x[i];
			x[j + 1] = x[i + 1];
			x[i] = rtemp;
			x[i + 1] = itemp;
		}
		for (m = N >> 1; m >= 2 && j >= m; m >>= 1)
			j -= m;
	}
}


/*
 * PHASE VOCODER INTERNALS 
 */

/*
 *    [was makewindows]
 *
 * make balanced pair of analysis (A) and synthesis (S) windows;
 * window lengths are Nw, FFT length is N, synthesis interpolation
 * factor is I
 *
 *
 * Call when fft_length, window_length, window_type are changed.
 */
static void rebuild_windows(GdamPhaseVocoder*      vocoder)
{
	int             i;
	double          sum;

	/* Analysis and synthesis windows. */
	gdouble*        A;
	gdouble*        S;

	GdamPhaseVocoderWindow window_type = vocoder->window_type;
	int window_length = vocoder->window_length;
	int fft_length = vocoder->fft_length;

	if (vocoder->analysis_window != NULL)
		g_free(vocoder->analysis_window);
	vocoder->analysis_window = A = g_new(gdouble, window_length);
	if (vocoder->synthesis_window != NULL)
		g_free(vocoder->synthesis_window);
	vocoder->synthesis_window = S = g_new(gdouble, window_length);

	if (window_type == GDAM_WINDOW_HAMMING) {
		for (i = 0; i < window_length; i++) {
			gdouble radians = TWOPI * i / (window_length - 1);
			A[i] = S[i] = 0.54 - 0.46 * cos(radians);
		}
	} else if (window_type == GDAM_WINDOW_HANNING) {
		for (i = 0; i < window_length; i++) {
			gdouble radians = TWOPI * i / (window_length - 1);
			A[i] = S[i] = 0.5 - 0.5 * cos(radians);
		}
	} else if (window_type == GDAM_WINDOW_RECT) {
		for (i = 0; i < window_length; i++) {
			A[i] = S[i] = 1.0;
		}
	} else if (window_type == GDAM_WINDOW_BARTLETT) {
		gdouble mid = (window_length - 1.0) / 2.0;
		for (i = 0; i < window_length; i++) {
			gdouble val;
			val = ((gdouble)i - mid);
			if (val < 0.0)
				val = (-val);
			A[i] = S[i] = 1.0 - val / mid;
		}
	} else {
		g_assert_not_reached();
	}

	/*
	 * when Nw > N, also apply interpolating (sinc) windows to
	 * ensure that window are 0 at increments of N (the FFT length)
	 * away from the center of the analysis window and of I away
	 * from the center of the synthesis window
	 */
	if (window_length > fft_length) {
		double          x;

		/*
		 * take care to create symmetrical windows
		 */
		x = -(window_length - 1) / 2.0;
		for (i = 0; i < window_length; i++, x += 1.0) {
			if (x != 0.0) {
				A[i] *= fft_length
				      * sin(PI * x / fft_length)
				      / (PI * x);
				S[i] *= I
				      * sin(PI * x / I)
				      / (PI * x);
			}
		}
	}

	/*
	 * normalize windows for unity gain across unmodified
	 * analysis-synthesis procedure
	 */
	sum = 0.0;
	for (i = 0; i < window_length; i++)
		sum += A[i];

	for (i = 0; i < window_length; i++) {
		gdouble analysis_factor, synthesis_factor;
		analysis_factor = 2.0 / sum;
		if (window_length > fft_length)
			synthesis_factor = 1.0 / analysis_factor;
		else
			synthesis_factor = analysis_factor;
		A[i] *= analysis_factor;
		S[i] *= synthesis_factor;
	}

	if (window_length <= fft_length) {
		for (sum = i = 0; i < window_length; i += I)
			sum += S[i] * S[i];
		for (sum = 1. / sum, i = 0; i < window_length; i++)
			S[i] *= sum;
	}
}

/* Shift data into the analysis windows. */
static int shiftin(GdamPhaseVocoder* vococder,
	           gdouble*          max_out,
		   GdamSample*       samples_in,
		   int               num_samples_in)
{
		   /*
                   gdouble* A,
                   int      fft_length,
	           int      D,
	           int      ch
		   */
	/* XXX: must iterate over left/right channels. */
	int i;
	gdouble *A_left = vocoder->analysis_window[0];
	gdouble* A_right = vocoder->analysis_window[1];
	int max_as_int = 0;
	int decimation = vocoder->decimation;

	if (si_valid < 0)
		si_valid = fft_length;
	i = fft_length - decimation;
	while (i--) {
		A_left[0] = A_left[decimation];
		A_right[0] = A_right[decimation];
		A_left++;
		A_right++;
	}

	/* XXX figure out what si_valid is supposed to do. */
	if (si_valid == fft_length) {
		short int *frames;
		int samples;
		GdamSample* samples_at;

		samples = num_samples_in;
		samples_at = samples_in;
		
		si_valid = sampsread + fft_length - decimation;
		while (sampsread--) {
			*A_left = SAMPLE_TO_FLOAT(samples_at->s[0]);
			*A_right = SAMPLE_TO_FLOAT(samples_at->s[1]);
			{
				int abs_left = ABS(samples_at->s[0]);
				int abs_right = ABS(samples_at->s[1]);
				if (abs_left > abs_right) {
					if (abs_left > max_as_int)
						abs_left = max_as_int;
				} else {
					if (abs_right > max_as_int)
						abs_right = max_as_int;
				}
			}
			A_left++;
			A_right++;
			samples_at++;
		}
	}
	if (si_valid < fft_length) {
		for (i = si_valid; i < fft_length; i++)
			A_left[i] = A_right[i] = 0.0;
		si_valid -= decimation;
	}
	*max_out = SAMPLE_TO_FLOAT(max_as_int);
	return (si_valid <= 0);
}


/*
 * multiply current input I by window W (both of length Nw);
 * using modulus arithmetic, fold and rotate windowed input
 * into output array O of (FFT) length fft_length according to current
 * input time n
 */
static void fold(gdouble*         I,
                 gdouble*         W,
		 int              Nw,
		 gdouble*         O,
		 int              fft_length,
		 int              n)
{

	int i;

	for (i = 0; i < N; i++)
		O[i] = 0.0;

	while (n < 0)
		n += N;
	n %= N;
	for (i = 0; i < Nw; i++) {
		O[n] += I[i] * W[i];
		if (++n == N)
			n = 0;
	}
}


/*
 * S is a spectrum in rfft format, i.e., it contains N real values arranged
 * as real followed by imaginary values, except for first two values, which
 * are real parts of 0 and Nyquist frequencies; convert first changes these
 * into N/2+1 PAIRS of magnitude and phase values to be stored in output
 * array C; the phases are then unwrapped and successive phase differences
 * are used to compute estimates of the instantaneous frequencies for each
 * phase vocoder analysis channel; decimation rate D and sampling rate R are
 * used to render these frequency values directly in Hz. 
 */

static void convert(gdouble*        S,
                    gdouble*        C,
		    int             D,
		    int             ch)
{
	double   fundamental, factor;
	double   phase, phasediff;
	int      real, imag, amp, freq;
	double   a, b;
	int      i;
	int      fft_length = vocoder->fft_length;
	int      half_fft_length = fft_length / 2;

	/*
	 * first pass: allocate zeroed space for previous phase values
	 *             for each channel and compute constants.
	 */

	fundamental = (double) SAMPLING_RATE / vocoder->fft_length;
	factor = SAMPLING_RATE / (decimation * TWOPI);

	/*
	 * Unravel rfft-format spectrum:
	 * Note that fft_length/2+1 pairs of values are produced.
	 */

	for (i = 0; i <= half_fft_length; i++) {
		imag = freq = (real = amp = i << 1) + 1;
		a = (i == half_fft_length ? S[1] : S[real]);
		b = (i == 0 || i == half_fft_length ? 0. : S[imag]);

		/*
		 * compute magnitude value from real and imaginary parts 
		 */

		C[amp] = hypot(a, b);

		/*
		 * compute phase value from real and imaginary parts and take
		 * difference between this and previous value for each channel 
		 */

		if (C[amp] == 0.0)
			phasediff = 0.0;
		else {
			phasediff = (phase = -atan2(b, a)) - lastphase[ch][i];
			lastphase[ch][i] = phase;

			/*
			 * unwrap phase differences 
			 */

			while (phasediff > PI)
				phasediff -= TWOPI;
			while (phasediff < -PI)
				phasediff += TWOPI;
		}

		/*
		 * convert each phase difference to Hz 
		 */

		C[freq] = phasediff * factor + i * fundamental;
	}
}

/*
 * oscillator bank resynthesizer for phase vocoder analyzer uses sum of N+1
 * cosinusoidal table lookup oscillators to compute I (interpolation factor) 
 * samples of output O from N+1 amplitude and frequency value-pairs in C;
 * frequencies are scaled by P 
 */

static void
oscbank (GdamPhaseVocoder* vocoder,
         gdouble           C[],
	 int               R,
	 int               I,
	 gdouble           O[],
	 int               ch,
	 gdouble           coef[],
	 gdouble           a0)
{
	static double  *table = NULL;
	int             amplitude, frequency, chan;
	double          Iinv;
	int             window_length = vocoder->window_length;
        int             fft_length = vocoder->fft_length;
	int             Np = vocoder->Np;

	/*
	 * First pass: allocate memory for and compute cosine table.
	 */

	if (table == NULL) {
		int n;
		table = g_new(gdouble, L);

		for (n = 0; n < L; n++)
			table[n] = N * cos(TWOPI * (double) n / L);
	}

	Iinv = 1.0 / I;

	/*
	 * For each channel, compute I samples using linear interpolation
	 * on the amplitude and frequency control values.
	 */

	for (chan = 0; chan < NP; chan++) {

		/* XXX: clean this up */
		double a, ainc, f, finc, address;

		frequency = (amplitude = (chan << 1)) + 1;
		C[frequency] *= Pinc;
		finc = (C[frequency] - (f = lastfreq[ch][chan])) * Iinv;
		if (C[amplitude] < synt)
			C[amplitude] = 0.;
		else if (Np)
			C[amplitude] *= lpamp(chan * P * PI / N, a0, coef, Np)
			        / lpamp(chan * PI / N, a0, coef, Np);
		ainc = (C[amplitude] - (a = lastamp[ch][chan])) * Iinv;
		address = index[ch][chan];


		/*
		 * Accumulate the I samples from each oscillator
		 * into output array O (initially assumed to be zero);
		 * f is frequency in Hz scaled by
		 * oscillator increment factor and pitch (Pinc);
		 * a is amplitude; 
		 */

		if (ainc != 0. || a != 0.) {
			int n;
			for (n = 0; n < I; n++) {
				O[n] += a * table[(int) address];

				address += f;

				while (address >= L)
					address -= L;

				while (address < 0)
					address += L;

				a += ainc;
				f += finc;
			}

			/*
			 * save current values for next iteration 
			 */
			lastfreq[ch][chan] = C[frequency];
			lastamp[ch][chan] = C[amplitude];
			index[ch][chan] = address;
		}
	}
}


/*
 * unconvert essentially undoes what convert does, i.e., it turns N2+1 PAIRS
 * of amplitude and frequency values in C into N2 PAIR of complex spectrum
 * data (in rfft format) in output array S; sampling rate R and interpolation 
 * factor I are used to recompute phase values from frequencies 
 */

static void
unconvert(gdouble    C[],
	  gdouble    S[],
	  int        N2,
	  int        I,
	  int        ch)
{
	int             i, real, imag, amplitude, frequency;
	double          magnitude, phase;

	/* These maybe should be cached in the vocoder (they were static) */
	gdouble         fundamental, factor;

	fundamental = (double) SAMPLING_RATE / (N2 << 1);
	factor = TWOPI * I / SAMPLING_RATE;

	/*
	 * subtract out frequencies associated with each channel,
	 * compute phases in terms of radians per I samples,
	 * and convert to complex form.
	 */

	for (i = 0; i <= N2; i++) {
		imag = frequency = (real = amplitude = i << 1) + 1;
		if (i == N2)
			real = 1;
		magnitude = C[amplitude];
		lastphase[ch][i] += C[frequency] - i * fundamental;
		phase = lastphase[ch][i] * factor;
		S[real] = magnitude * cos(phase);
		if (i != N2)
			S[imag] = -magnitude * sin(phase);
	}
}


/*
 * input I is a folded spectrum of length N; output O and
 * synthesis window W are of length Nw--overlap-add windowed,
 * unrotated, unfolded input data into output O
 */
static void
overlapadd(gdouble   I[],
	   int       fft_length,
	   gdouble   W[],
	   gdouble   O[],
	   int       window_length)
{
	int i;
	while (n < 0)
		n += fft_length;
	n %= N;
	for (i = 0; i < window_length; i++) {
		O[i] += I[n] * W[i];
		if (++n == fft_length)
			n = 0;
	}
}

#define FLOAT_TO_SAMPLE(value)	((GdamSampleType)((value) * GDAM_SAMPLE_MAX))

void
shiftout(GdamPhaseVocoder* vocoder, int I, int n)
{
	int i, ch;

	gdouble** output = vocoder->output;
	int fft_length = vocoder->fft_length;

	/* XXX: I and n probably don't need to be parameters. */

	if (n >= 0) {
		for (i = 0; i < I; i++) {
			framebuf[i * NUM_CHANNELS + 0]
				= FLOAT_TO_SAMPLE(output[0][i]);
			framebuf[i * NUM_CHANNELS + 1]
				= FLOAT_TO_SAMPLE(output[1][i]);
		}
		AFwriteframes(outfile, AF_DEFAULT_TRACK, framebuff, I);
	}
	for (ch = 0; ch < samps_per_frame; ch++) {
		for (i = 0; i < fft_length - I; i++)
			output[ch][i] = output[ch][i + I];
		for (i = fft_length - I; i < fft_length; i++)
			output[ch][i] = 0.;
	}
}


/*                      rightbins
 * 
 * A function to shimmy the amplitude and frequency arrays
 * so that they are close to k / 44100 hertz,
 * where `k' is the array index.
 */
void
rightbins()
{
	double          nyquist;
	long            point = 0;
	int             i, j, thebin;

	nyquist = R / 2.;
	for (i = 0; i < horiz_num * samps_per_frame; i++) {
		for (j = 0; j < num_pv_channels; j++)
			tempamp[j] = 0.;
		for (j = 0; j < num_pv_channels; j++) {
			gdouble frequency_dom = megfreq[point + j];
			gdouble amplitude_dom = megamp[point + j];
			if (frequency_dom > 0.0 && frequency_dom < nyquist) {
				thebin = (int) (0.5 + frequency_dom / binfreq);
				if (thebin >= num_pv_channels)
					g_warning("PV: OOPS: %d", thebin);
				if (amplitude_dom > tempamp[thebin]) {
					tempamp[thebin] = amplitude_dom;
					tempfreq[thebin] = frequency_dom;
				}
			}
		}
		for (j = 0; j < num_pv_channels; j++) {
			megfreq[point + j] = tempfreq[j];
			megamp[point + j] = tempamp[j];
		}
		point += num_pv_channels;
	}
}

/*
 * Remove frequencies that are invalid or above the nyquist rate
 * (which will just come out as noise).
 */
static void
clipfreq()
{
	double          nyquist;
	long            point = 0;
	int             i, j, thebin;

	nyquist = R / 2.;
	for (i = 0; i < horiz_num * samps_per_frame; i++) {
		for (j = 0; j < num_pv_channels; j++) {
			if (megfreq[point] < 0.0 || megfreq[point] >= nyquist)
				megamp[point] = 0.0;
			point++;
		}
	}
}

/* XXX: what the hell does this do? */
double
getfuncval(int hnum)
{
	double          hnumx;

	hnumx = hnum * 800. / horiz_num + 50.;
	while ((int) floor(hnumx) > xpos[funcpt + 1])
		funcpt++;
	return theheight - (ypos[funcpt + 1] - ypos[funcpt]) * (hnumx -
							    xpos[funcpt]) /
		(xpos[funcpt + 1] - xpos[funcpt]) - ypos[funcpt] + 10;

}


/*
 * CALLBACK ROUTINES 
 */

void
loadana(char *filename)
{
    int             in,
                    i,
                    ch,
                    eof,
                    grafx,
                    grafold,
                    j,
                    y;
    long            point,
                    p;
    double          amp;
    XPoint          list0[LSIZE],
                    list1[LSIZE],
                    list2[LSIZE],
                    list3[LSIZE];
    int             c0 = 0,
                    c1 = 0,
                    c2 = 0,
                    c3 = 0;
    double          a;

	init_again();
	horiz_num = (framecnt / Dn) + overlap;

	if (num_pv_channels * horiz_num * samps_per_frame > 32000000) {
		XtManageChild(sizeWarning);
		AFclosefile(infile);
		return;
	}

	//duration = (double) framecnt / R;
	binfreq = (double) R / N;
	if (megfreq != NULL)
		free(megfreq);
	if (megamp != NULL)
		free(megamp);
	if (largamp != NULL)
		free(largamp);

	/* These used to contain the entire sample data / and fft'd data. */
	//megfreq = megamp = NULL;
	//largamp = NULL;
}


static void SieveOk()
{
	int             numparint, numparint2, i, j, k, maxnum, ch,
                        pr, lastpr = -1;
        long            point;
        double         *tempamp, *megamp2, max, mult_pr_horiz, numpar,
                        sievew, rscale;
        gboolean        amplitude_sel, function_sel, null_sel;

    fvec(tempamp, num_pv_channels);
    fvec(megamp2, num_pv_channels);

	    /* Parameters... */
	    numparint2 = atoi(val1);
	    mult_pr_horiz = powf(atof(val2), duration / horiz_num);
	    sievew = atof(val3);

	    amplitude_sel = XmToggleButtonGetState(sieveToggle1);
	    function_sel = XmToggleButtonGetState(sieveToggle3);
	    null_sel = XmToggleButtonGetState(sieveToggle4);

    for (ch = 0; ch < samps_per_frame; ch++) {
	funcpt = 1;
	maxnum = 1;
	numparint = numparint2;
	numpar = (double) numparint;
	rscale = (double) (areaf2 - areaf1) / 32768.;
	point = ch * num_pv_channels * horiz_num + areat1 * num_pv_channels;
	for (i = areat1; i < areat2; i++) {
	    if (funcsel)
		numparint = (int) getfuncval(i);
	    if (!amplitude_sel && (numparint > num_pv_channels - 10))
		numparint = num_pv_channels - 10;
	    for (j = 0; j < num_pv_channels; j++) {
		if (nullsel)
		    tempamp[j] = megamp[point + j];
		else
		    tempamp[j] = 0.;
		megamp2[j] = megamp[point + j] * (1. + sievew * j);
	    }
	    if (ampsel)
		for (k = 0; k < numparint; k++) {
		    max = 0.;
		    if (ampsel)
			for (j = areaf1; j < areaf2; j++) {
			    if (megamp2[j] >= max) {
				max = megamp2[j];
				maxnum = j;
			    }
			}
		    if (nullsel)
			tempamp[maxnum] = 0.;
		    else
			tempamp[maxnum] = megamp[point + maxnum];
		    megamp2[maxnum] = -1.;
	    } else
		for (k = 0; k < numparint; k++) {
		    j = 0;
		    do {
			maxnum = (int) ((double) rand() * rscale) + areaf1;
			j++;
		    }
		    while ((megamp2[maxnum] == -1) && (j < num_pv_channels));
		    if (nullsel)
			tempamp[maxnum] = 0.;
		    else
			tempamp[maxnum] = megamp[point + maxnum];
		    megamp2[maxnum] = -1.;
		}
	    for (j = areaf1; j < areaf2; j++)
		megamp[point + j] = tempamp[j];
	    point += num_pv_channels;
	    if (!funcsel) {
		numpar *= mult_pr_horiz;
		numparint = (int) (floor(numpar));
	    }
	    pr = (int) (100. * (double) (i - areat1) / (areat2 - areat1));
	    if (pr != lastpr) {
		XtVaSetValues(progressScale, XmNvalue, pr, NULL);
		XmUpdateDisplay(topLevel);
		lastpr = -pr;
	    }
	}
    }
}

void
SshiftOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    int             i,
                    j,
                    ch;
    long            point;
    double          shval,
                    shval2,
                    mult_pr_horiz;
    Boolean         funcsel;

    valstring = XmTextGetString(sshiftTextfield);
    shval2 = atof(valstring);
    XtFree(valstring);
    valstring = XmTextGetString(sshiftTextfield2);
    mult_pr_horiz = powf(atof(valstring), duration / horiz_num);
    XtFree(valstring);
    funcsel = XmToggleButtonGetState(sshiftToggle1);

    XmUpdateDisplay(w);
    SetWatchCursor(topLevel);

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * horiz_num * num_pv_channels;
	funcpt = 1;
	shval = shval2;
	for (i = areat1; i < areat2; i++) {
	    if (funcsel)
		shval =
		    1000. * (getfuncval(i) / (double) theheight) - 500.;
	    for (j = areaf1; j < areaf2; j++) {
		megfreq[point + j] += shval;
		if (megfreq[point + j] < 20.)
		    megamp[point + j] *= megfreq[point + j] / 20.;
	    }
	    point += num_pv_channels;
	    shval *= mult_pr_horiz;
	}
    }
    rightbins();
    RedrawWin();
    ResetCursor(topLevel);
}

void
ShearOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    int             i,
                    j,
                    ch;
    long            delay,
                    moveto,
                    start;
    double          shval,
                    shval2,
                    mult_pr_horiz,
                    fact;
    Boolean         funcsel;

    valstring = XmTextGetString(shearTextfield1);
    shval2 = atof(valstring);
    XtFree(valstring);
    valstring = XmTextGetString(shearTextfield2);
    mult_pr_horiz = powf(atof(valstring), duration / horiz_num);
    XtFree(valstring);
    funcsel = XmToggleButtonGetState(shearToggle1);

    XmUpdateDisplay(w);
    SetWatchCursor(topLevel);

    for (ch = 0; ch < samps_per_frame; ch++) {
	funcpt = 1;
	start = ch * num_pv_channels * horiz_num;
	if (shval2 < 0)
	    for (i = areat1; i < areat2; i++) {
		shval = -shval2 * powf(mult_pr_horiz, i);
		if (funcsel)
		    shval = getfuncval(i) / (double) theheight;
		fact = R * shval / (1000. * Dn);
		for (j = areaf1; j < areaf2; j++) {
		    delay =
			(long) (fact *
				megfreq[i * num_pv_channels + j + start]);
		    moveto = i - delay;
		    if (moveto >= horiz_num)
			moveto = horiz_num - 1;
		    if (moveto < 0)
			moveto = 0;
		    megfreq[moveto * num_pv_channels + j + start] =
			megfreq[i * num_pv_channels + j + start];
		    megamp[moveto * num_pv_channels + j + start] =
			megamp[i * num_pv_channels + j + start];
		    megfreq[i * num_pv_channels + j + start] = 0.;
		    megamp[i * num_pv_channels + j + start] = 0.;
		}
	} else
	    for (i = areat2 - 1; i >= areat1; i--) {
		shval = shval2 * powf(mult_pr_horiz, i - areat1);
		if (funcsel)
		    shval = getfuncval(i) / (double) theheight;
		fact = R * shval / (1000. * Dn);
		for (j = areaf1; j < areaf2; j++) {
		    delay =
			(long) (fact *
				megfreq[i * num_pv_channels + j + start]);
		    moveto = i + delay;
		    if (moveto >= horiz_num)
			moveto = horiz_num - 1;
		    if (moveto < 0)
			moveto = 0;
		    megfreq[moveto * num_pv_channels + j + start] =
			megfreq[i * num_pv_channels + j + start];
		    megamp[moveto * num_pv_channels + j + start] =
			megamp[i * num_pv_channels + j + start];
		    megfreq[i * num_pv_channels + j + start] = 0.;
		    megamp[i * num_pv_channels + j + start] = 0.;
		}
	    }
    }
    RedrawWin();
    ResetCursor(topLevel);
}

void
BlurOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    int             i,
                    j,
                    p,
                    ch,
                    stepn;
    long            start,
                    t;
    double          blurval,
                    blurval2,
                    mult_pr_horiz,
                    ya,
                    yf,
                    step;
    Boolean         funcsel;

    valstring = XmTextGetString(blurTextfield1);
    blurval2 = atof(valstring);
    XtFree(valstring);
    valstring = XmTextGetString(blurTextfield2);
    mult_pr_horiz = powf(atof(valstring), duration / horiz_num);
    XtFree(valstring);
    valstring = XmTextGetString(blurTextfield3);
    step = atof(valstring);
    if (step == 0.)
	step = 1e-9;
    stepn = (int) ((R / (double) Dn) / step);
    XtFree(valstring);
    funcsel = XmToggleButtonGetState(blurToggle1);

    XmUpdateDisplay(w);
    SetWatchCursor(topLevel);

    for (ch = 0; ch < samps_per_frame; ch++) {
	start = ch * num_pv_channels * horiz_num;
	for (j = areaf1; j < areaf2; j++) {
	    t = 0;
	    blurval = blurval2;
	    funcpt = 1;
	    if (blurval > 0.)
		for (i = areat1; i < areat2; i++) {
		    p = i * num_pv_channels + j + start;
		    if (funcsel)
			blurval = getfuncval(i) / (double) theheight;
		    if (blurval > 1.0)
			blurval = 1.0;
		    if (t % stepn == 0) {
			ya = megamp[p];
			yf = megfreq[p];
		    } else {
			ya = blurval * ya + (1. - blurval) * megamp[p];
			megamp[p] = ya;
			yf = blurval * yf + (1. - blurval) * megfreq[p];
			megfreq[p] = yf;
		    }
		    blurval *= mult_pr_horiz;
		    t++;
	    } else
		for (i = areat1 + 1; i < areat2; i++) {
		    p = i * num_pv_channels + j + start;
		    if (funcsel)
			blurval = getfuncval(i) / (double) theheight;
		    if (blurval < -1.0)
			blurval = -1.0;
		    if (t % stepn == 0) {
			ya = megamp[p];
			yf = megfreq[p];
		    } else {
			ya = blurval * ya + (1. + blurval) * megamp[p];
			megamp[p] = ya;
		    }
		    blurval *= mult_pr_horiz;
		    t++;
		}
	}
    }
    rightbins();
    RedrawWin();
    ResetCursor(topLevel);
}

void
CombOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    int             i,
                    j,
                    ch,
                    bin,
                    basebin;
    long            point;
    double          combval,
                    combval2,
                    mult_pr_horiz;
    Boolean         funcsel;

    valstring = XmTextGetString(combTextfield1);
    combval2 = atof(valstring);
    XtFree(valstring);
    valstring = XmTextGetString(combTextfield2);
    mult_pr_horiz = powf(atof(valstring), duration / horiz_num);
    XtFree(valstring);
    funcsel = XmToggleButtonGetState(combToggle1);

    XmUpdateDisplay(w);
    SetWatchCursor(topLevel);

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = ch * num_pv_channels * horiz_num + areat1 * num_pv_channels;
	funcpt = 1;
	combval = combval2;
	for (i = areat1; i < areat2; i++) {
	    for (j = areaf1; j < areaf2; j++) {
		if (funcsel)
		    combval = 1000. * getfuncval(i) / (double) theheight;
		basebin = (int) (combval / binfreq + 0.5);
		if (basebin >= num_pv_channels)
		    basebin = num_pv_channels - 1;
		if (basebin) {
		    if (j % basebin)
			megamp[point] = 0.;
		}
		combval *= mult_pr_horiz;
		point++;
	    }
	}
    }
    RedrawWin();
    ResetCursor(topLevel);
}

void
MirrorOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    int             i,
                    j,
                    ch;
    long            point;
    double          mval,
                    mval2,
                    mult_pr_horiz,
                    f;
    Boolean         funcsel;

    valstring = XmTextGetString(mirrorTextfield);
    mval2 = atof(valstring);
    XtFree(valstring);
    valstring = XmTextGetString(mirrorTextfield2);
    mult_pr_horiz = powf(atof(valstring), duration / horiz_num);
    XtFree(valstring);
    funcsel = XmToggleButtonGetState(mirrorToggle);

    XmUpdateDisplay(w);
    SetWatchCursor(topLevel);

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * horiz_num * num_pv_channels;
	funcpt = 1;
	mval = mval2;
	for (i = areat1; i < areat2; i++) {
	    if (funcsel)
		mval = binfreq * getfuncval(i);
	    for (j = areaf1; j < areaf2; j++) {
		f = megfreq[point + j];
		f = mval - (f - mval);
		if (f < 0.)
		    f = 0.;
		if (f < 20.)
		    f *= f / 20.;
		megfreq[point + j] = f;
	    }
	    point += num_pv_channels;
	    mval *= mult_pr_horiz;
	}
    }
    rightbins();
    RedrawWin();
    ResetCursor(topLevel);
}

void
ExpoOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    int             i,
                    j,
                    ch;
    long            point;
    double          eval,
                    eval2,
                    mult_pr_horiz,
                    f,
                    mul;
    Boolean         funcsel;

    valstring = XmTextGetString(expoTextfield);
    eval2 = atof(valstring);
    XtFree(valstring);
    valstring = XmTextGetString(expoTextfield2);
    mult_pr_horiz = powf(atof(valstring), duration / horiz_num);
    XtFree(valstring);
    valstring = XmTextGetString(expoTextfield3);
    mul = atof(valstring);
    XtFree(valstring);
    funcsel = XmToggleButtonGetState(expoToggle);

    XmUpdateDisplay(w);
    SetWatchCursor(topLevel);

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	funcpt = 1;
	eval = eval2;
	for (i = areat1; i < areat2; i++) {
	    if (funcsel)
		eval = 2. * getfuncval(i) / (double) theheight;
	    for (j = areaf1; j < areaf2; j++) {
		f = megfreq[point + j];
		f = powf(f * mul, eval);
		if (f < 0.)
		    f = 0.;
		if (f < 20.)
		    f *= f / 20.;
		megfreq[point + j] = f;
	    }
	    point += num_pv_channels;
	    eval *= mult_pr_horiz;
	}
    }
    rightbins();
    RedrawWin();
    ResetCursor(topLevel);
}

void
SpreadOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    int             i,
                    j,
                    ch;
    long            point;
    double          eval,
                    eval2,
                    mult_pr_horiz,
                    f,
                    mul;
    Boolean         funcsel;

    valstring = XmTextGetString(spreadTextfield);
    eval2 = atof(valstring);
    XtFree(valstring);
    valstring = XmTextGetString(spreadTextfield2);
    mult_pr_horiz = powf(atof(valstring), duration / horiz_num);
    XtFree(valstring);
    valstring = XmTextGetString(spreadTextfield3);
    mul = atof(valstring);
    XtFree(valstring);
    funcsel = XmToggleButtonGetState(spreadToggle);

    XmUpdateDisplay(w);
    SetWatchCursor(topLevel);

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	funcpt = 1;
	eval = eval2;
	for (i = areat1; i < areat2; i++) {
	    if (funcsel)
		eval = 2. * getfuncval(i) / (double) theheight;
	    for (j = areaf1; j < areaf2; j++) {
		f = megfreq[point + j];
		f *=
		    powf(2.,
			 eval * (rand() / 16384. - 1.) * (1. - mul +
							  mul *
							  megamp[point +
								 j] *
							  10000.));
		if (f < 0.)
		    f = 0.;
		if (f < 20.)
		    f *= f / 20.;
		megfreq[point + j] = f;
	    }
	    point += num_pv_channels;
	    eval *= mult_pr_horiz;
	}
    }
    rightbins();
    RedrawWin();
    ResetCursor(topLevel);
}

void
PshiftOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    int             i,
                    j,
                    ch;
    long            point;
    double          fact,
                    fact2,
                    mult_pr_horiz,
                    reference,
                    diff;
    Boolean         funcsel;

    valstring = XmTextGetString(pitchTextfield);
    fact2 = atof(valstring);
    XtFree(valstring);
    valstring = XmTextGetString(pitchTextfield2);
    mult_pr_horiz = powf(atof(valstring), duration / horiz_num);
    XtFree(valstring);
    valstring = XmTextGetString(pitchTextfield3);
    reference = atof(valstring);
    funcsel = XmToggleButtonGetState(pitchToggle1);

    XmUpdateDisplay(w);
    SetWatchCursor(topLevel);

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	funcpt = 1;
	fact = fact2;
	for (i = areat1; i < areat2; i++) {
	    if (funcsel)
		fact = 2. * getfuncval(i) / (double) theheight;
	    for (j = areaf1; j < areaf2; j++) {
		diff = megfreq[point + j] - reference;
		megfreq[point + j] = fact * diff + reference;
		if (megfreq[point + j] < 20.)
		    megamp[point + j] *= megfreq[point + j] / 20.;
	    }
	    point += num_pv_channels;
	    fact *= mult_pr_horiz;
	}
    }
    rightbins();
}


void
filterOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    long            point,
                    ch;
    int             i,
                    j,
                    k,
                    filter_lower,
                    filter_upper,
                    ant,
                    keep;

    if (XmToggleButtonGetState(filterFilToggle)) {
	for (ch = 0; ch < samps_per_frame; ch++) {
	    point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	    ant = makefilterfile();
	    if (ant > 0)
		for (i = areat1; i < areat2; i++) {
		    for (j = areaf1; j < areaf2; j++) {
			keep = 0;
			for (k = 0; k < ant; k++) {
			    filter_lower = filterfreqs[k] - binfreq / 2.;
			    filter_upper = filterfreqs[k] + binfreq / 2.;
			    if ((megfreq[point + j] >= filter_lower)
				&& (megfreq[point + j] <= filter_upper))
				keep = 1;
			}
			if (keep == 0)
			    megamp[point + j] = 0.;
		    }
		    point += num_pv_channels;
		}
	}
    } else {
	valstring = XmTextGetString(filterTextfield);
	filter_lower = atoi(valstring);
	XtFree(valstring);
	if (filter_lower == 0.)
	    filter_lower = -99999.;
	valstring = XmTextGetString(filterTextfield2);
	filter_upper = atoi(valstring);
	XtFree(valstring);

	for (ch = 0; ch < samps_per_frame; ch++) {
	    point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	    for (i = areat1; i < areat2; i++) {
		for (j = areaf1; j < areaf2; j++)
		    if ((megfreq[point + j] >= filter_lower)
			&& (megfreq[point + j] <= filter_upper))
			megamp[point + j] = 0.;
		point += num_pv_channels;
	    }
	}
	XtFree(valstring);
    }

    RedrawWin();
}

void
GainOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    double          val;
    long            point;
    int             i,
                    j,
                    ch;
    Boolean         funcsel;

    funcsel = XmToggleButtonGetState(gainToggle);

    valstring = XmTextGetString(gainTextfield);
    val = atof(valstring);
    XtFree(valstring);

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	for (i = areat1; i < areat2; i++) {
	    if (funcsel)
		val = getfuncval(i) / (double) theheight;
	    for (j = areaf1; j < areaf2; j++)
		megamp[point + j] *= val;
	    point += num_pv_channels;
	}
    }

    XtFree(valstring);
}

void
ThresholdOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    double          val,
                    maxamp;
    long            point;
    int             i,
                    j,
                    ch;
    Boolean         funcsel;

    funcsel = XmToggleButtonGetState(thresholdToggle);

    valstring = XmTextGetString(thresholdTextfield);
    val = atof(valstring);
    XtFree(valstring);

    maxamp = -1.;
    for (ch = 0; ch < samps_per_frame; ch++) {
	point = ch * num_pv_channels * horiz_num;
	for (i = 0; i < horiz_num; i++) {
	    for (j = 0; j < num_pv_channels; j++)
		if (megamp[point + j] > maxamp)
		    maxamp = megamp[point + j];
	    point += num_pv_channels;
	}
    }

    val *= maxamp;
    if (maxamp > 0.)
	for (ch = 0; ch < samps_per_frame; ch++) {
	    point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	    for (i = areat1; i < areat2; i++) {
		if (funcsel)
		    val = maxamp * getfuncval(i) / (double) theheight;
		for (j = areaf1; j < areaf2; j++)
		    if (megamp[point + j] < val)
			megamp[point + j] = 0.;
		point += num_pv_channels;
	    }
	}

    RedrawWin();
    XtFree(valstring);
}

void
envelopenoise(ampvec)
     double         *ampvec;
{
    int             i,
                    j;
    double          max;

    for (i = 1; i < num_pv_channels / 8 - 1; i++) {
	max = 0.;
	for (j = 0; j < 8; j++)
	    if (ampvec[i * 8 + j] > max)
		max = ampvec[i * 8 + j];
	for (j = 0; j < 8; j++)
	    ampvec[i * 8 + j] = max;
    }
}

void
NoiseredOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    double          val,
                    maxamp,
                   *tempamp;
    long            point;
    int             i,
                    j,
                    ch;
    Boolean         funcsel,
                    envelopesel,
                    despecklesel;

    funcsel = XmToggleButtonGetState(noiseredToggle);
    envelopesel = XmToggleButtonGetState(noiseredToggle2);
    despecklesel = XmToggleButtonGetState(noiseredToggle3);

    valstring = XmTextGetString(noiseredTextfield);
    val = atof(valstring);
    XtFree(valstring);

    fvec(tempamp, num_pv_channels);

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	for (j = 0; j < num_pv_channels; j++)
	    tempamp[j] = 0.;
	for (i = areat1; i < areat2; i++) {
	    for (j = 0; j < num_pv_channels; j++)
		tempamp[j] += megamp[point + j];
	    point += num_pv_channels;
	}
	for (j = 0; j < num_pv_channels; j++)
	    tempamp[j] /= (areat2 - areat1);
	if (envelopesel)
	    envelopenoise(tempamp);
	point = ch * num_pv_channels * horiz_num;
	for (i = 0; i < horiz_num; i++) {
	    if (funcsel)
		val = 2. * getfuncval(i) / (double) theheight;
	    for (j = areaf1; j < areaf2; j++) {
		megamp[point + j] -= val * tempamp[j];
		if (megamp[point + j] < 0.)
		    megamp[point + j] = 0.;
	    }
	    point += num_pv_channels;
	}
	if (despecklesel) {
	    point = ch * num_pv_channels * horiz_num + num_pv_channels;
	    for (i = 1; i < horiz_num - 1; i++) {
		for (j = areaf1; j < areaf2; j++) {
		    if ((megamp[point + j - num_pv_channels] == 0.) &&
			(megamp[point + j + num_pv_channels] == 0.))
			megamp[point + j] = 0.;
		}
		point += num_pv_channels;
	    }
	    point = ch * num_pv_channels * horiz_num;
	    for (i = 0; i < horiz_num; i++) {
		for (j = areaf1 + 1; j < areaf2 - 1; j++) {
		    if ((megamp[point + j - 1] == 0.) &&
			(megamp[point + j + 1] == 0.))
			megamp[point + j] = 0.;
		}
		point += num_pv_channels;
	    }
	    point = ch * num_pv_channels * horiz_num + num_pv_channels;
	    for (i = 1; i < horiz_num - 1; i++) {
		for (j = areaf1; j < areaf2; j++) {
		    if ((megamp[point + j - num_pv_channels] == 0.) &&
			(megamp[point + j + num_pv_channels] == 0.))
			megamp[point + j] = 0.;
		}
		point += num_pv_channels;
	    }
	}
    }

    RedrawWin();
    XtFree(valstring);
    free(tempamp);
}

void
Nullphase(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    long            point;
    int             i,
                    j,
                    ch;

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	for (i = areat1; i < areat2; i++) {
	    for (j = areaf1; j < areaf2; j++)
		megfreq[point + j] = j * binfreq;
	    point += num_pv_channels;
	}
    }
    RedrawWin();
}

void
Trpeaks(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    long            point;
    int             i,
                    j,
                    keep,
                    ch;
    double         *tempamp;

    fvec(tempamp, num_pv_channels);

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * horiz_num * num_pv_channels;
	for (i = areat1; i < areat2; i++) {
	    for (j = areaf1; j < areaf2; j++)
		tempamp[j] = megamp[point + j];
	    for (j = areaf1; j < areaf2 - 1; j++) {
		keep = 0;
		if (j == areaf1)
		    keep = 1;
		else {
		    if ((tempamp[j - 1] < tempamp[j])
			&& (tempamp[j + 1] < tempamp[j]))
			keep = 1;
		}
		if ((tempamp[j] <= 0.000001) || (megfreq[point + j] < 20.)
		    || (keep == 0))
		    megamp[point + j] = 0.;
	    }
	    point += num_pv_channels;
	}
    }
    free(tempamp);
    RedrawWin();
}

void
LTAS(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    int             i,
                    j,
                    ch,
                    start;
    double          middleamp,
                    middlefreq;

    for (ch = 0; ch < samps_per_frame; ch++) {
	start = ch * horiz_num * num_pv_channels;
	for (j = 0; j < num_pv_channels; j++) {
	    middleamp = middlefreq = 0.;
	    for (i = 0; i < horiz_num; i++) {
		middleamp += megamp[i * num_pv_channels + j + start];
		middlefreq += megfreq[i * num_pv_channels + j + start];
	    }
	    middleamp /= horiz_num;
	    middlefreq /= horiz_num;
	    for (i = 0; i < horiz_num; i++) {
		megamp[i * num_pv_channels + j + start] = middleamp;
		megfreq[i * num_pv_channels + j + start] = middlefreq;
	    }
	}
    }
    RedrawWin();
}

void
makepenta()
{
    char           *valstring;
    float           val;
    int             i;

    valstring = XmTextGetString(gridTextfield);
    val = atof(valstring);
    XtFree(valstring);

    gridfreqs[0] = val;
    gridfreqs[1] = gridfreqs[0] * 1.122;
    gridfreqs[2] = gridfreqs[0] * 1.335;
    gridfreqs[3] = gridfreqs[0] * 1.498;
    gridfreqs[4] = gridfreqs[0] * 1.682;
    for (i = 5; i < 500; i++) {
	gridfreqs[i] = gridfreqs[i - 5] * 2.;
	numfreqs = i + 1;
	if (gridfreqs[i] > 23000)
	    break;
    }
}

void
makemajor(void)
{
    char           *valstring;
    float           val;
    int             i;

    valstring = XmTextGetString(gridTextfield);
    val = atof(valstring);
    XtFree(valstring);

    gridfreqs[0] = val;
    gridfreqs[1] = gridfreqs[0] * 1.122;
    gridfreqs[2] = gridfreqs[0] * 1.260;
    gridfreqs[3] = gridfreqs[0] * 1.335;
    gridfreqs[4] = gridfreqs[0] * 1.498;
    gridfreqs[5] = gridfreqs[0] * 1.682;
    gridfreqs[6] = gridfreqs[0] * 1.888;
    for (i = 7; i < 500; i++) {
	gridfreqs[i] = gridfreqs[i - 7] * 2.;
	numfreqs = i + 1;
	if (gridfreqs[i] > 23000)
	    break;
    }
}

void
makeminor(void)
{
    char           *valstring;
    float           val;
    int             i;

    valstring = XmTextGetString(gridTextfield);
    val = atof(valstring);
    XtFree(valstring);

    gridfreqs[0] = val;
    gridfreqs[1] = gridfreqs[0] * 1.122;
    gridfreqs[2] = gridfreqs[0] * 1.189;
    gridfreqs[3] = gridfreqs[0] * 1.335;
    gridfreqs[4] = gridfreqs[0] * 1.498;
    gridfreqs[5] = gridfreqs[0] * 1.587;
    gridfreqs[6] = gridfreqs[0] * 1.888;
    for (i = 7; i < 500; i++) {
	gridfreqs[i] = gridfreqs[i - 7] * 2.;
	numfreqs = i + 1;
	if (gridfreqs[i] > 23000)
	    break;
    }
}

void
makechroma(void)
{
    char           *valstring;
    float           val;
    int             i;

    valstring = XmTextGetString(gridTextfield);
    val = atof(valstring);
    XtFree(valstring);

    gridfreqs[0] = val;
    gridfreqs[1] = gridfreqs[0] * 1.0594631;
    gridfreqs[2] = gridfreqs[0] * 1.1224621;
    gridfreqs[3] = gridfreqs[0] * 1.189;
    gridfreqs[4] = gridfreqs[0] * 1.260;
    gridfreqs[5] = gridfreqs[0] * 1.335;
    gridfreqs[6] = gridfreqs[0] * 1.414;
    gridfreqs[7] = gridfreqs[0] * 1.498;
    gridfreqs[8] = gridfreqs[0] * 1.587;
    gridfreqs[9] = gridfreqs[0] * 1.682;
    gridfreqs[10] = gridfreqs[0] * 1.782;
    gridfreqs[11] = gridfreqs[0] * 1.888;
    for (i = 12; i < 500; i++) {
	gridfreqs[i] = gridfreqs[i - 12] * 2;
	numfreqs = i + 1;
	if (gridfreqs[i] > 23000)
	    break;
    }
}

void
makewhole(void)
{
    char           *valstring;
    float           val;
    int             i;

    valstring = XmTextGetString(gridTextfield);
    val = atof(valstring);
    XtFree(valstring);

    gridfreqs[0] = val;
    gridfreqs[1] = gridfreqs[0] * 1.1224621;
    gridfreqs[2] = gridfreqs[0] * 1.260;
    gridfreqs[3] = gridfreqs[0] * 1.414;
    gridfreqs[4] = gridfreqs[0] * 1.587;
    gridfreqs[5] = gridfreqs[0] * 1.782;
    for (i = 6; i < 500; i++) {
	gridfreqs[i] = gridfreqs[i - 6] * 2;
	numfreqs = i + 1;
	if (gridfreqs[i] > 23000)
	    break;
    }
}

void
makeharmseries(void)
{
    char           *valstring;
    float           val;
    int             i;

    valstring = XmTextGetString(gridTextfield);
    val = atof(valstring);
    XtFree(valstring);

    gridfreqs[0] = val;
    numfreqs = 1;
    for (i = 1; i < 500; i++) {
	gridfreqs[i] = gridfreqs[i - 1] + val;
	numfreqs = i + 1;
	if (gridfreqs[i] > 23000)
	    break;
    }
}

void
makeevenharmseries(void)
{
    char           *valstring;
    float           val;
    int             i;

    valstring = XmTextGetString(gridTextfield);
    val = atof(valstring);
    XtFree(valstring);

    gridfreqs[0] = val;
    gridfreqs[1] = 2 * val;
    for (i = 2; i < 500; i++) {
	gridfreqs[i] = gridfreqs[i - 1] + (2 * val);
	numfreqs = i + 1;
	if (gridfreqs[i] > 23000)
	    break;
    }
}

void
makeoddharmseries(void)
{
    char           *valstring;
    float           val;
    int             i;

    valstring = XmTextGetString(gridTextfield);
    val = atof(valstring);
    XtFree(valstring);

    gridfreqs[0] = val;
    for (i = 1; i < 500; i++) {
	gridfreqs[i] = gridfreqs[i - 1] + (2 * val);
	numfreqs = i + 1;
	if (gridfreqs[i] > 23000)
	    break;
    }
}

void
gridsort(double *gridfreqs2, int left, int right)
{
    int             i,
                    j;
    float           x,
                    y;

    i = left;
    j = right;
    x = gridfreqs2[(left + right) / 2];

    do {
	while (gridfreqs2[i] < x && i < right)
	    i++;
	while (x < gridfreqs2[j] && j > left)
	    j--;

	if (i <= j) {
	    y = gridfreqs2[i];
	    gridfreqs2[i] = gridfreqs2[j];
	    gridfreqs2[j] = y;
	    i++;
	    j--;
	}
    } while (i <= j);

    if (left < j)
	gridsort(gridfreqs2, left, j);
    if (i < right)
	gridsort(gridfreqs2, i, right);
}


void
MagnetOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    int             i,
                    j,
                    k,
                    nearest,
                    ch,
                    totf;
    double          mindiff,
                    f,
                    proba,
                    proba2,
                    mult_pr_horiz;
    long            point = 0;
    char           *valstring;
    Boolean         funcsel;

    valstring = XmTextGetString(magnetTextfield);
    proba2 = atof(valstring);
    XtFree(valstring);
    valstring = XmTextGetString(magnetTextfield2);
    mult_pr_horiz = powf(atof(valstring), duration / horiz_num);
    XtFree(valstring);
    funcsel = XmToggleButtonGetState(magnetToggle1);

    SetWatchCursor(topLevel);

    totf = numfreqs;
    for (i = 0; i < totf; i++)
	gridfreqs2[i] = gridfreqs[i];

    gridsort(gridfreqs2, 0, totf - 1);

    for (i = 0; i < totf; i++) {
	while (gridfreqs2[i] == gridfreqs2[i + 1]) {
	    for (j = i; j < numfreqs - 1; j++)
		gridfreqs2[j] = gridfreqs2[j + 1];
	    totf--;
	}
    }

    for (ch = 0; ch < samps_per_frame; ch++) {
	point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	funcpt = 1;
	proba = proba2;
	for (i = areat1; i < areat2; i++) {
	    if (funcsel)
		proba = getfuncval(i) / (double) theheight;
	    for (j = areaf1; j < areaf2; j++) {
		if ((double) rand() / 32768. < proba) {
		    nearest = totf - 1;
		    mindiff = 99999.;
		    for (k = 0; k < 100; k++) {
			f = fabs(megfreq[point + j] - gridfreqs2[k]);
			if (f < mindiff) {
			    nearest = k;
			    mindiff = f;
			} else
			    break;
		    }
		    megfreq[point + j] = gridfreqs2[nearest];
		    if (megfreq[point + j] < 20.)
			megamp[point + j] *= megfreq[point + j] / 20.;
		}
	    }
	    point += num_pv_channels;
	    proba *= mult_pr_horiz;
	}
    }
    rightbins();
    RedrawWin();
    ResetCursor(topLevel);
}

void
ConvolveOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    double          laplace[3][3] = {
	{0, -1, 0},
	{-1, 4, -1},
	{0, -1, 0}
    };
    double   mean[3][3] = {
	{0.11, 0.11, 0.11},
	{0.11, 0.11, 0.11},
	{0.11, 0.11, 0.11}
    };
    double          timemean[3][3] = {
	{0., 0., 0.},
	{0.33, 0.33, 0.33},
	{0., 0., 0.}
    };
    double          op[3][3];

    int             i,
                    j,
                    i1,
                    j1,
                    ch;
    long            point;
    double          sum;

    XmUpdateDisplay(w);
    SetWatchCursor(topLevel);

    for (i = 0; i <= 2; i++)
	for (j = 0; j <= 2; j++) {
	    if (XmToggleButtonGetState(convolveToggleLaplace))
		op[i][j] = laplace[j][i];
	    if (XmToggleButtonGetState(convolveToggleMean))
		op[i][j] = mean[j][i];
	    if (XmToggleButtonGetState(convolveToggleTimeMean))
		op[i][j] = timemean[j][i];
	}

    for (ch = 0; ch < samps_per_frame; ch++) {
	for (j = 0; j < num_pv_channels; j++)
	    this[j] = 0.;
	point = areat1 * num_pv_channels + ch * num_pv_channels * horiz_num;
	for (i = areat1; i < areat2; i++) {
	    for (j = 0; j < num_pv_channels; j++)
		last[j] = this[j];
	    for (j = 0; j < num_pv_channels; j++)
		this[j] = megamp[point + j];
	    for (j = areaf1; j < areaf2; j++) {
		sum = 0.;
		for (j1 = -1; j1 <= 1; j1++) {
		    if ((j + j1 >= 0) && (j + j1 < num_pv_channels))
			sum += op[0][j1 + 1] * last[j + j1];
		}
		for (j1 = -1; j1 <= 1; j1++) {
		    if ((j + j1 >= 0) && (j + j1 < num_pv_channels))
			sum += op[1][j1 + 1] * this[j + j1];
		}
		for (j1 = -1; j1 <= 1; j1++) {
		    if ((i + 1 < horiz_num) && (j + j1 >= 0)
			&& (j + j1 < num_pv_channels)) sum +=
			    op[2][j1 + 1] * megamp[point + num_pv_channels +
						   j + j1];
		}
		if (sum >= 0.)
		    megamp[point + j] = sum;
	    }
	    point += num_pv_channels;
	}
    }
    RedrawWin();
    ResetCursor(topLevel);
}

void
TocsoundOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    XmFileSelectionBoxCallbackStruct
	* calldata = (XmFileSelectionBoxCallbackStruct *) call;
    char           *filename;
    static XmStringCharSet charset =
	(XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
    FILE           *scorefile;
    int             i,
                    j;
    long            point;
    double          distance = stretchfac * duration / (double) horiz_num;
    double          dur = stretchfac / binfreq;
    double          onset;

    XmStringGetLtoR(calldata->value, charset, &filename);
    scorefile = fopen(filename, "w");
    if (scorefile == NULL) {
	XtManageChild(csoundWarning);
	XtFree(filename);
	return;
    }

    point = areat1 * num_pv_channels;
    for (i = areat1; i < areat2; i++) {
	onset = i * distance;
	for (j = areaf1; j < areaf2; j++) {
	    if ((megamp[point + j] > 0.) && (megfreq[point + j] > 20.))
		fprintf(scorefile,
			"i1    %8.4f    %8.4f    %7.2f    %7.2f\n", onset,
			dur, megamp[point + j] * 10000000.,
			megfreq[point + j]);
	}
	point += num_pv_channels;
    }

    fprintf(scorefile, "e\n");

    fclose(scorefile);
    XtFree(filename);
    XtUnmanageChild(tocsoundFileBox);
}

void
ToparmerudOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    XmFileSelectionBoxCallbackStruct
	* calldata = (XmFileSelectionBoxCallbackStruct *) call;
    char           *filename,
                   *valstring;
    static XmStringCharSet charset =
	(XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
    FILE           *scorefile;
    int             i,
                    j,
                    k,
                    strings,
                    state,
                    maxpoint;
    long            point = 0;
    double          distance = stretchfac * duration / (double) horiz_num;
    double          dur = stretchfac / binfreq;
    double          onset,
                    saveamp;
    double          triglevel,
                    gatetime,
                    resontime,
                    strfrq,
                    range,
                    maxamp,
                    timeren;

    XmStringGetLtoR(calldata->value, charset, &filename);
    scorefile = fopen(filename, "w");
    if (scorefile == NULL) {
	XtManageChild(csoundWarning);
	return;
    }

    valstring = XmTextGetString(parmT1);
    triglevel = atof(valstring);
    valstring = XmTextGetString(parmT2);
    gatetime = atof(valstring);
    valstring = XmTextGetString(parmT3);
    strings = atoi(valstring);
    valstring = XmTextGetString(parmT4);
    resontime = atof(valstring);

    strfrq = 30.;
    for (i = 0; i < strings * 8; i++) {
	state = 0;
	range = (strfrq * powf(2., (double) strings / 12.) - strfrq) / 2.;
	for (j = 0; j < horiz_num; j++) {
	    point = j * num_pv_channels;
	    maxamp = -100.;
	    maxpoint = -1;
	    for (k = 0; k < num_pv_channels; k++) {
		if ((megfreq[point] > strfrq - range)
		    && (megfreq[point] < strfrq + range)
		    && (megamp[point] * 10000000. > triglevel)) {
		    if (megamp[point] > maxamp) {
			maxamp = megamp[point];
			maxpoint = point;
		    }
		}
		point++;
	    }
	    if (state == 0) {
		if (maxpoint > -1) {
		    saveamp = maxamp * 10000000.;
		    onset = j * distance;
		    state = 1;
		}
	    } else if (state == 1) {
		if (maxpoint == -1) {
		    fprintf(scorefile,
			    "i1    %8.4f    %8.4f    %7.2f    %7.2f\n",
			    onset,
			    (resontime >
			     j * distance -
			     onset ? resontime : j * distance - onset),
			    saveamp, strfrq);
		    state = 2;
		    timeren = 0.;
		}
	    } else {
		timeren += distance;
		if (timeren > gatetime)
		    state = 0;
	    }
	}
	strfrq *= powf(2., 1. / (double) strings);
    }

    fprintf(scorefile, "e\n");

    fclose(scorefile);
    XtFree(filename);
    XtUnmanageChild(toparmerudFileBox);
}

void
TopeakOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    XmFileSelectionBoxCallbackStruct
	* calldata = (XmFileSelectionBoxCallbackStruct *) call;
    char           *filename;
    static XmStringCharSet charset =
	(XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
    FILE           *scorefile;
    int             i,
                    j,
                    keep;
    long            point;
    double          distance = stretchfac * duration / (double) horiz_num;
    double          dur = stretchfac / binfreq;
    double          onset;

    XmStringGetLtoR(calldata->value, charset, &filename);
    scorefile = fopen(filename, "w");
    if (scorefile == NULL) {
	XtManageChild(csoundWarning);
	XtFree(filename);
	return;
    }

    point = areat1 * num_pv_channels;
    for (i = areat1; i < areat2; i++) {
	onset = i * distance;
	for (j = 0; j < num_pv_channels; j++) {
	    keep = 0;
	    if (j == 0)
		keep = 1;
	    else {
		if ((megamp[point - 1] < megamp[point])
		    && (megamp[point + 1] < megamp[point]))
		    keep = 1;
	    }
	    if ((megamp[point] > 0.000001) && (megfreq[point] > 20.)
		&& keep) fprintf(scorefile, "%8.4f    %7.2f    %7.2f\n",
				 onset, megamp[point] * 10000000.,
				 megfreq[point]);
	    point++;
	}
    }

    fclose(scorefile);
    XtFree(filename);
    XtUnmanageChild(topeakFileBox);
}


void
ToltasOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    XmFileSelectionBoxCallbackStruct
	* calldata = (XmFileSelectionBoxCallbackStruct *) call;
    char           *filename;
    static XmStringCharSet charset =
	(XmStringCharSet) XmSTRING_DEFAULT_CHARSET;
    FILE           *ltasfile;
    int             i,
                    j;
    long            point;

    XmStringGetLtoR(calldata->value, charset, &filename);
    ltasfile = fopen(filename, "w");
    if (ltasfile == NULL) {
	XtManageChild(ltasWarning);
	XtFree(filename);
	return;
    }

    point = 4 * num_pv_channels;
    for (j = 0; j < num_pv_channels; j++)
	fprintf(ltasfile, "%f\n", megamp[point++] * 10000);

    fclose(ltasfile);
    XtFree(filename);
    XtUnmanageChild(toltasFileBox);
}


void
AnalyOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    int             i;
    char           *valstring;
    double         *A = Wanal,
                   *S = Wsyn;

    valstring = XmTextGetString(analyTextfield);
    overlap = atoi(valstring);
    XtFree(valstring);
    if (overlap < 1)
	overlap = 1;
    if (overlap > 32)
	overlap = 32;

    Dn = Nw / overlap;
    I = Dn;
    makewindows(Wanal, Wsyn, Nw, N, I);

    valstring = XmTextGetString(analyTextfield2);
    numcoef = atoi(valstring);
    XtFree(valstring);
    if (numcoef < 1)
	numcoef = 1;
    if (numcoef > 100)
	numcoef = 100;
    if (numcoef != numcoef2)
	numcoef2 = 0;
}

void
GridmenuOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    if (XmToggleButtonGetState(gridTogglePenta))
	makepenta();
    else if (XmToggleButtonGetState(gridToggleMajor))
	makemajor();
    else if (XmToggleButtonGetState(gridToggleMinor))
	makeminor();
    else if (XmToggleButtonGetState(gridToggleChroma))
	makechroma();
    else if (XmToggleButtonGetState(gridToggleWhole))
	makewhole();
    else if (XmToggleButtonGetState(gridToggleHarm))
	makeharmseries();
    else if (XmToggleButtonGetState(gridToggleEvenharm))
	makeevenharmseries();
    else if (XmToggleButtonGetState(gridToggleOddharm))
	makeoddharmseries();
    else
	PGFreqfile(NULL, NULL, NULL);
    RestoreWin(NULL, NULL, NULL);
}

void
SpotsRestart(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    numspots = 0;
    RedrawWin();
}

void
SpotsKeep(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    int             i,
                    j,
                    k,
                    f,
                    t,
                    ch;
    long            start;

    for (ch = 0; ch < samps_per_frame; ch++) {
	start = ch * num_pv_channels * horiz_num;
	for (i = 0; i < numspots; i++) {
	    for (j = -3; j < 4; j++) {
		f = spotsf[i] + j;
		if ((f > 0) && (f < num_pv_channels)) {
		    for (k = -3 * horiz_num / 800.;
			 k < 4 * horiz_num / 800.; k++) {
			t = spotst[i] + k;
			if ((t >= 0) && (t < horiz_num))
			    megamp[t * num_pv_channels + f + start] =
				-megamp[t * num_pv_channels + f + start];
		    }
		}
	    }
	}

	for (i = 0; i < horiz_num; i++) {
	    for (j = 0; j < num_pv_channels; j++) {
		megamp[start] = (megamp[start] < 0 ? -megamp[start] : 0.);
		start++;
	    }
	}
    }
    RedrawWin();
}

void
SpotsRemove(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    int             i,
                    j,
                    k,
                    f,
                    t,
                    ch;

    for (ch = 0; ch < samps_per_frame; ch++)
	for (i = 0; i < numspots; i++) {
	    for (j = -3; j < 4; j++) {
		f = spotsf[i] + j;
		if ((f > 0) && (f < num_pv_channels)) {
		    for (k = -3 * horiz_num / 800.;
			 k < 4 * horiz_num / 800.; k++) {
			t = spotst[i] + k;
			if ((t >= 0) && (t < horiz_num))
			    megamp[ch * horiz_num * num_pv_channels +
				   t * num_pv_channels + f] = 0.;
		    }
		}
	    }
	}

    RedrawWin();
}

void
Amplitude()
{
	double          amp;
	int             j, lastx, lasty, antnod = 1, x, lastxx = -1;

	for (j = 0; j < horiz_num; j++) {
		x = (int) (j * 800. / horiz_num) + 50;
		if (x != lastxx) {
			amp = largamp[j];
			if (amp > 1.)
				amp = 1.;
			amp = (amp / 2.) + 0.25;
			amp = (1. - amp) * theheight + 10;
			if (j == 0) {
				lastx = 50;
				lasty = (int) amp;
				xpos[1] = 50;
				ypos[1] = (int) amp;
			} else if ((antnod < MAXNOD)
				   && ((fabs(amp - lasty) > 50.)
				       || (x - lastx > 20))) {
				antnod++;
				lastx = x;
				lasty = (int) amp;
				xpos[antnod] = x;
				ypos[antnod] = (int) amp;
			}
			lastxx = x;
		}
	}
	xpos[antnod] = 850;
	numsquare = antnod;
}

void
Pitch(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    double          maxamp,
                    maxamp2,
                    freq,
                    freq2;
    int             i,
                    j,
                    point = 0,
                    lastx,
                    lasty,
                    antnod = 1,
                    x,
                    lastxx = -1,
                    maxbin,
                    maxbin2;

    for (j = 0; j < horiz_num; j++) {
	x = (int) (j * 800. / horiz_num) + 50;
	if (x != lastxx) {

	    maxamp = 0.;
	    maxbin = 1;
	    for (i = 2; i < 40; i++)
		if (megamp[point + i] > maxamp) {
		    maxamp = megamp[point + i];
		    maxbin = i;
		}
	    freq = megfreq[maxbin + point];
	    freq2 = 0.;
	    if (freq < 60.)
		freq = 0.;
	    else {
		maxamp2 = 0.;
		maxbin2 = 1;
		for (i = 2; i < maxbin - 1; i++)
		    if (megamp[point + i] > maxamp2) {
			maxamp2 = megamp[point + i];
			maxbin2 = i;
		    }
		freq2 = megfreq[maxbin2 + point];
		if ((fabs(freq / freq2 - 2.) < 0.1)
		    && (maxamp / maxamp2 < 16.)) freq = freq2;
		else if ((fabs(freq / freq2 - 3.) < 0.1)
			 && (maxamp / maxamp2 < 16.))
		    freq = freq2;
		else if ((fabs(freq / freq2 - 1.5) < 0.1)
			 && (maxamp / maxamp2 < 16.))
		    freq = freq2 / 2.;
	    }

	    freq = freq / 1000.;
	    if (freq > 1.)
		freq = 1.;
	    freq = (1. - freq) * theheight + 10.;

	    if (j == 0) {
		lastxx = 50;
		lasty = (int) freq;
		xpos[1] = 50;
		ypos[1] = (int) freq;
	    }
		else if ((antnod < MAXNOD)
			 && ((fabs(freq - lasty) > 50.)
			     || (x - lastx > 20))) {
		antnod++;
		lastx = x;
		lasty = (int) freq;
		xpos[antnod] = x;
		ypos[antnod] = (int) freq;
	    }
	    lastxx = x;
	}
	point += num_pv_channels;
    }
    xpos[antnod] = 850;
    numsquare = antnod;
}

void
Centroid(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    double          centro,
                    norm;
    int             i,
                    j,
                    point = 0,
                    lastx,
                    lasty,
                    antnod = 1,
                    x,
                    lastxx = -1;

    for (j = 0; j < horiz_num; j++) {
	x = (int) (j * 800. / horiz_num) + 50;
	if (x != lastxx) {
	    centro = 0.;
	    norm = 0.;
	    for (i = 1; i < num_pv_channels - 1; i++) {
		centro += megamp[point + i] * megfreq[point + i];
		norm += megamp[point + i];
	    }
	    centro = (norm <= 0. ? 0. : centro / norm);
	    /*
	     * if (max<0.01) centro=0.; 
	     */
	    centro /= 5000.;
	    if (centro > 1.)
		centro = 1.;
	    centro = (centro / 2.) + 0.25;
	    centro = (1. - centro) * theheight + 10.;

	    if (j == 0) {
		lastxx = 50;
		lasty = (int) centro;
		xpos[1] = 50;
		ypos[1] = (int) centro;
	    }
		else if ((antnod < MAXNOD)
			 && ((fabs(centro - lasty) > 50.)
			     || (x - lastx > 20))) {
		antnod++;
		lastx = x;
		lasty = (int) centro;
		xpos[antnod] = x;
		ypos[antnod] = (int) centro;
	    }
	    lastxx = x;
	}
	point += num_pv_channels;
    }
    xpos[antnod] = 850;
    numsquare = antnod;
}

void
GrScaleOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    double          val,
                    y;
    int             i;

    valstring = XmTextGetString(grScaleTextfield);
    val = atof(valstring);
    XtFree(valstring);

    for (i = 1; i <= numsquare; i++) {
	y = val * (1. - (ypos[i] - 10.) / theheight);
	if (y < 0.)
	    y = 0.;
	if (y > 1.)
	    y = 1.;
	ypos[i] = (1. - y) * theheight + 10;
    }
}

void
GrMoveOk(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    char           *valstring;
    double          val,
                    y;
    int             i;

    valstring = XmTextGetString(grMoveTextfield);
    val = atof(valstring);
    XtFree(valstring);

    for (i = 1; i <= numsquare; i++) {
	y = val + (1. - (ypos[i] - 10.) / theheight);
	if (y < 0.)
	    y = 0.;
	if (y > 1.)
	    y = 1.;
	ypos[i] = (1. - y) * theheight + 10;
    }
}

void
GrNonlin(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
}

void
GrClear(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    ypos[1] = 10 + theheight / 2;
    ypos[2] = ypos[1];
    xpos[2] = 850;
    numsquare = 2;
    selected = 0;
}

void
GrInvert(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
    double          y;
    int             i;

    for (i = 1; i <= numsquare; i++) {
	y = 1. - (1. - (ypos[i] - 10.) / theheight);
	if (y < 0.)
	    y = 0.;
	if (y > 1.)
	    y = 1.;
	ypos[i] = (1. - y) * theheight + 10;
    }
}

void
GrRetro(w, client, call)
     Widget          w;
     XtPointer       client,
                     call;
{
}

void
gdam_phase_vocoder_init(GdamPhaseVocoder * phase_vocoder)
{
	int num_pv_channels;
	/*
	 * INITIALIZE PHASE VOCODER 
	 */

	phase_vocoder->fft_length = 1024;
	phase_vocoder->overlap = 8;

	/*
	* Window size 
	*/
	phase_vocoder->window_length = phase_vocoder->fft_length;

	/*
	* Decimation factor 
	*/
	phase_vocoder->decimation = phase_vocoder->window_length
	                          / phase_vocoder->overlap;
	/*
	* Interpolation factor 
	*/
	phase_vocoder->I = phase_vocoder->decimation;
	// N2=N>>1; theheight=N2;

	/*
	* Oscbank pitch factor 
	*/
	phase_vocoder->P = phase_vocoder->1.0;

	phase_vocoder->Pinc = phase_vocoder->P * 8192.0 / R;
	phase_vocoder->NP = phase_vocoder->fft_length / 2;


	/*
	* Synthesis threshold 
	*/
	phase_vocoder->synt = 0.0;

	/*
	* ... 
	*/
	phase_vocoder->binfreq = (double) R / phase_vocoder->fft_length;
	phase_vocoder->obank = 0.;
	//Nw2 = Nw >> 1;
	num_pv_channels = phase_vocoder->fft_length / 2 + 1;
	phase_vocoder->num_pv_channels = num_pv_channels;

	phase_vocoder->Wanal = new_double_array(phase_vocoder->window_length);
	phase_vocoder->Wsyn = new_double_array(phase_vocoder->window_length);
	phase_vocoder->input = new_double_array(phase_vocoder->window_length);
	phase_vocoder->buffer = new_double_array(phase_vocoder->fft_length);
	phase_vocoder->channel = new_double_array(phase_vocoder->fft_length+2);

	phase_vocoder->tempamp = new_double_array(num_pv_channels);
	phase_vocoder->tempfreq = new_double_array(num_pv_channels);
	phase_vocoder->last = new_double_array(num_pv_channels);
	phase_vocoder->this = new_double_array(num_pv_channels);

}
