/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdampluckedstring.h"
#include <math.h>
#include <stdlib.h>

static GtkObjectClass* parent_class;

static gint16 clamp(gint sample)
{
	return (sample < GDAM_SAMPLE_MIN) ? GDAM_SAMPLE_MIN
	    :  (sample > GDAM_SAMPLE_MAX) ? GDAM_SAMPLE_MAX
	    :  sample;
}
static int gdam_plucked_string_read(GdamSource* source, GdamSample* buf, int n)
{
	GdamSource* filtered = GDAM_FILTER(source)->filtered;
	GdamPluckedString* plucked = GDAM_PLUCKED_STRING(source);
	GdamCircularBuffer* lowpass = &plucked->lowpass;
	GdamFixed a, gain;
	int num_read;
	GdamFixed feedback_div_4 = plucked->gain / 4;
	GdamSample last_allpass_out = plucked->last_allpass_out;
	GdamSample last_allpass_in = plucked->last_allpass_in;
	if (filtered == NULL)
		return 0;
	num_read = n = gdam_source_read(filtered, buf, n);
	if (n <= 0)
		return n;

	g_return_val_if_fail(lowpass->amount_buffered >= 2, -1);

	a = plucked->a;
	gain = plucked->gain;

	while (n-- > 0) {
		int next_out[2];
		GdamSample lowpass_output, lowpass_input, tmp;

		lowpass_output = gdam_circular_buffer_read_one(lowpass);
		tmp = gdam_circular_buffer_peek_one(lowpass);
		if (plucked->filter_coeff != GdamFixed_ONE) {
			tmp.s[0] = (int)(plucked->filter_coeff
			                          * tmp.s[0]
			                          / GdamFixed_ONE);
			tmp.s[1] = (int)(plucked->filter_coeff
			                          * tmp.s[1]
			                          / GdamFixed_ONE);
		}
		lowpass_output.s[0] = ((int)tmp.s[0] + lowpass_output.s[0]) / 2;
		lowpass_output.s[1] = ((int)tmp.s[1] + lowpass_output.s[1]) / 2;

		next_out[0] =  (int)(last_allpass_in.s[0])
		            + (( - (gint64)(last_allpass_out.s[0] * a)
		                 + (gint64)(lowpass_output.s[0] * a)) >> 16);
		next_out[1] =  (int)(last_allpass_in.s[1])
		            + (( - (gint64)(last_allpass_out.s[1] * a)
		                 + (gint64)(lowpass_output.s[1] * a)) >> 16);

		last_allpass_in = lowpass_output;
		lowpass_input.s[0] = clamp(next_out[0] * feedback_div_4
		   					/ (GdamFixed_ONE/4)
					      + buf->s[0]);
		lowpass_input.s[1] = clamp(next_out[1] * feedback_div_4
		   					/ (GdamFixed_ONE/4)
					      + buf->s[1]);
		gdam_circular_buffer_append_one(lowpass, lowpass_input);
		buf->s[0] = clamp(next_out[0]);
		buf->s[1] = clamp(next_out[1]);
		last_allpass_out = *buf;
		buf++;
	}
	plucked->last_allpass_out = last_allpass_out;
	plucked->last_allpass_in = last_allpass_in;
	return num_read;
}
static void gdam_plucked_string_destroy(GtkObject* object)
{
	GdamPluckedString* plucked = GDAM_PLUCKED_STRING(object);
	gdam_circular_buffer_destruct(&plucked->lowpass);
	(*parent_class->destroy)(object);
}

static void gdam_plucked_string_arg_set_hertz(GtkObject* object, GdamArg* arg)
{
	gdam_plucked_string_set_hertz(GDAM_PLUCKED_STRING(object),
	                              GDAM_ARG_DOUBLE(arg));
}

static void gdam_plucked_string_arg_get_hertz(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_DOUBLE(arg) = GDAM_PLUCKED_STRING(object)->hertz;
}

static void gdam_plucked_string_arg_set_gain(GtkObject* object, GdamArg* arg)
{
	gdam_plucked_string_set_gain(GDAM_PLUCKED_STRING(object),
	                             GDAM_ARG_DOUBLE(arg));
}

static void gdam_plucked_string_arg_get_gain(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_DOUBLE(arg) = (gdouble)GDAM_PLUCKED_STRING(object)->gain
						/ GdamFixed_ONE;
}

static void gdam_plucked_string_arg_set_filter_coeff
					(GtkObject* object, GdamArg* arg)
{
	GdamFixed coeff = (int) (GDAM_ARG_DOUBLE(arg) * GdamFixed_ONE);
	GDAM_PLUCKED_STRING(object)->filter_coeff = coeff;
}

static void gdam_plucked_string_arg_get_filter_coeff
                                        (GtkObject* object, GdamArg* arg)
{
	gdouble coeff = GDAM_PLUCKED_STRING(object)->filter_coeff;
	coeff /= GdamFixed_ONE;
	GDAM_ARG_DOUBLE(arg) = coeff;
}

static void gdam_plucked_string_arg_do_pluck(GtkObject* object, GdamArg* arg)
{
	gdam_plucked_string_pluck(GDAM_PLUCKED_STRING(object),
	                             GDAM_ARG_DOUBLE(arg));
}

static void gdam_plucked_string_init(GdamPluckedString* plucked_string)
{
	plucked_string->filter_coeff = GdamFixed_ONE;
}
static void gdam_plucked_string_class_init(GdamPluckedStringClass* klass)
{
	GDAM_SOURCE_CLASS(klass)->read = gdam_plucked_string_read;
	GTK_OBJECT_CLASS(klass)->destroy = gdam_plucked_string_destroy;
	gdam_arg_register(GTK_OBJECT_CLASS(klass),
	                  "hertz",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_plucked_string_arg_set_hertz,
			  gdam_plucked_string_arg_get_hertz,
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(klass),
	                  "gain",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_plucked_string_arg_set_gain,
			  gdam_plucked_string_arg_get_gain,
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(klass),
	                  "filter_coeff",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_plucked_string_arg_set_filter_coeff,
			  gdam_plucked_string_arg_get_filter_coeff,
			  GDAM_ARG_VITAL);
	gdam_arg_register(GTK_OBJECT_CLASS(klass),
	                  "pluck",
			  GDAM_ARG_TYPE_DOUBLE,
			  gdam_plucked_string_arg_do_pluck,
			  NULL,
			  0);
}

GtkType gdam_plucked_string_get_type()
{
	static GtkType plucked_string_type = 0;
	if (!plucked_string_type) {
		static const GtkTypeInfo plucked_string_info = {
			"GdamPluckedString",
			sizeof(GdamPluckedString),
			sizeof(GdamPluckedStringClass),
			(GtkClassInitFunc) gdam_plucked_string_class_init,
			(GtkObjectInitFunc) gdam_plucked_string_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_FILTER;
		parent_class = gtk_type_class(parent);
		plucked_string_type = gtk_type_unique(parent,
		                                 &plucked_string_info);
	}
	return plucked_string_type;
}

void gdam_plucked_string_set_hertz(GdamPluckedString*  string,
                                   gdouble             hertz)
{
	gdouble period = 44100.0 / hertz;
	int lowpass_len = floor(period + 0.5);
	gdouble allpass_delay = period - ((gdouble)lowpass_len - 0.5);
	gdouble half_frequency = 0.5 / period;
	gdouble alpha;
	g_assert(0.0 <= allpass_delay && allpass_delay < 1.0);
	alpha = sin((1.0 - allpass_delay) * half_frequency)
	      / sin((1.0 + allpass_delay) * half_frequency);
	string->a = (GdamFixed)(alpha * GdamFixed_ONE);
	string->hertz = 440.0;
	gdam_circular_buffer_resize(&string->lowpass, lowpass_len, TRUE);
}

void gdam_plucked_string_set_gain (GdamPluckedString*  string,
                                   gdouble             gain)
{
	string->gain = (GdamFixed) (gain * GdamFixed_ONE);
}

static void add_noise(GdamSample* buf, int len, gdouble velo)
{
	int range = 65534.0 * velo;
	if (range == 0)
		return;
	while (len-- > 0) {
		int add = rand() % range - range/2;
		buf->s[0] += add;
		buf->s[1] += add;			/* TODO: clamp */
		buf++;
	}
}

void gdam_plucked_string_pluck           (GdamPluckedString*  filter,
                                          gdouble             velocity)
{
	GdamCircularBuffer* buffer = &filter->lowpass;
	add_noise(buffer->buffer, buffer->buffer_size, velocity);
}
