/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include "gdambuffereditorskin.h"
#include <gdam/gdam-protocol.h>
#include <gtk/gtksignal.h>
#include "../gdamskinutil.h"
#include <gdam/component/gdambuffermanager.h>
#include "../../component/plugins/gdamproducercache.h"
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtklabel.h>
static gboolean       gdam_buffer_editor_build_new_model
                                                 (GdamSkin*      skin,
                                                  GdamChannel*   channel,
						  int		 num_options,
						  char**         options);

static gboolean gdam_buffer_editor_viewport_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamBufferEditorSkin* editor = GDAM_BUFFER_EDITOR_SKIN(skin);
	if (!GTK_IS_CONTAINER(object)) {
		g_warning("cannot add buffer_viewport to non-container");
		return TRUE;
	}
	if (!editor->viewport){
		g_warning("cannot add non-existant buffer_viewport");
		return TRUE;
	}
	gtk_container_add(GTK_CONTAINER(object),
	                  GDAM_SKIN(editor->viewport)->main_widget);
	return TRUE;
}
static void zoom_changed(GtkAdjustment*      adjustment,
                         GdamBufferViewport* viewport)
{
	gdouble zoom = exp(adjustment->value);
	if (zoom > 10000.0) {
		return;
	}
	gdam_buffer_viewport_set_zoom(viewport, zoom);
	return;
}
static void viewport_zoom_changed(GdamBufferViewport* viewport,
				  GtkAdjustment*      adjustment)
{
	gdouble zoom = viewport->time_length;
	if (zoom < 0.0001)
		return;
	gtk_adjustment_set_value(adjustment, log(zoom));
}
static void connect_zoom_handler(GdamBufferViewport* viewport_skin,
                                 GtkAdjustment* adjustment)
{
	GtkObject* adj = GTK_OBJECT(adjustment);
	GtkObject* skin = GTK_OBJECT(viewport_skin);
	gtk_signal_connect(adj,
	                   "value_changed",
			   GTK_SIGNAL_FUNC(zoom_changed),
			   skin);
	gtk_signal_connect_while_alive 
	                  (skin,
	                   "view_changed",
			   GTK_SIGNAL_FUNC(viewport_zoom_changed),
			   adj,
	                   adj);
	gdam_delayed_adjustment_init(adjustment);
}

static gboolean gdam_buffer_editor_log_zoom_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj = util_get_widget_adjustment(GTK_WIDGET(object));
	GdamBufferEditorSkin* editor_skin = GDAM_BUFFER_EDITOR_SKIN(skin);
	if (editor_skin->viewport == NULL)
		return FALSE;
	connect_zoom_handler(editor_skin->viewport, adj);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static void center_changed(GtkAdjustment*      adjustment,
                         GdamBufferViewport* viewport)
{
	gdam_buffer_viewport_set_center(viewport, adjustment->value);
}
static void viewport_center_changed(GdamBufferViewport* viewport,
				    GtkAdjustment*      adjustment)
{
	if (viewport->use_buffer_manager){
		return;
	}
	adjustment->page_size = viewport->time_length;
	adjustment->page_increment = viewport->time_length / 10.0;
	adjustment->step_increment = viewport->time_length / 100.0;
	adjustment->upper = viewport->sample_length + adjustment->page_size;
	adjustment->value = viewport->time_start + viewport->time_length / 2;
	gtk_adjustment_changed(adjustment);
}

static void connect_center_handler(gpointer viewport_skin, gpointer adjustment)
{
	GtkObject* adj = GTK_OBJECT(adjustment);
	GtkObject* skin = GTK_OBJECT(viewport_skin);
	gtk_signal_connect(adj,
	                   "value_changed",
			   GTK_SIGNAL_FUNC(center_changed),
			   skin);
	gtk_signal_connect_while_alive
	                  (skin,
			   "length_changed",
			   GTK_SIGNAL_FUNC(viewport_center_changed),
			   adj,
	                   adj);
	gtk_signal_connect_while_alive
	                  (skin,
			   "view_changed",
			   GTK_SIGNAL_FUNC(viewport_center_changed),
			   adj,
	                   adj);
}

static void on_buffer_id_received(
			    const char*       text,
			    gpointer          data)
{
	GdamBufferEditorSkin* skin = GDAM_BUFFER_EDITOR_SKIN(data);
	int id = atoi(text);
	gdam_debug("buffer id %d dropped", id);
	gdam_buffer_viewport_set_buffer_id(skin->viewport, id);
}

static gboolean gdam_buffer_editor_buffer_id_dropped_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gdam_widget_string_dest_set(GTK_WIDGET(object),
				    on_buffer_id_received,
				    "gdam-buffer-id-text",
				    skin);
	return TRUE;
}



static gboolean gdam_buffer_editor_center_adjustment_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GtkAdjustment* adj = util_get_widget_adjustment(GTK_WIDGET(object));
	GdamBufferEditorSkin* editor_skin = GDAM_BUFFER_EDITOR_SKIN(skin);
	if (editor_skin->viewport == NULL)
		return FALSE;
	connect_center_handler(editor_skin->viewport, adj);
	gdam_delayed_adjustment_init(adj);
	return TRUE;
}

static void do_set_defaults(GdamBufferEditorSkin* editor_skin)
{
	int id;
	if (editor_skin->viewport == NULL)
		return;
	id = gdam_buffer_manager_obtain_id(NULL);
	gdam_buffer_manager_unref(id, NULL);
	gdam_buffer_viewport_set_buffer_id(editor_skin->viewport, id);
}

static gboolean gdam_buffer_editor_set_default_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          "clicked", 
				  GTK_SIGNAL_FUNC(do_set_defaults),
				  (GtkObject*) skin);
	return TRUE;
}

static void zoom_to_selection(GdamBufferEditorSkin* editor_skin)
{
	GdamBufferViewport* viewport = editor_skin->viewport;
	if (viewport == NULL)
		return;
	if (viewport->has_selection) {
		viewport->time_start = viewport->selection_start;
		viewport->time_length = viewport->selection_length;
		gdam_buffer_viewport_changed(viewport);
	}
}

static gboolean gdam_buffer_editor_zoom_to_selection_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamBufferEditorSkin* editor_skin = GDAM_BUFFER_EDITOR_SKIN(skin);
	gtk_signal_connect_object(object,
	                          signal_name,
                                  GTK_SIGNAL_FUNC(zoom_to_selection),
				  (GtkObject*) editor_skin);
	return TRUE;
}

static void zoom_to_sample(GdamBufferEditorSkin* editor_skin)
{
	GdamBufferViewport* viewport = editor_skin->viewport;
	if (viewport == NULL)
		return;
	gdam_buffer_viewport_whole_sample(viewport);
}

static gboolean gdam_buffer_editor_zoom_to_sample_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamBufferEditorSkin* editor_skin = GDAM_BUFFER_EDITOR_SKIN(skin);
	gtk_signal_connect_object(object,
	                          signal_name,
                                  GTK_SIGNAL_FUNC(zoom_to_sample),
				  (GtkObject*) editor_skin);
	return TRUE;
}

static void play_callback(GdamBufferEditorSkin* editor_skin)
{
	GdamProducerCache* cache;
	GdamSkin* skin = GDAM_SKIN(editor_skin);
	GdamBufferViewport* viewport = editor_skin->viewport;
	int samps_per_sec;
	g_return_if_fail(viewport != NULL);
	g_return_if_fail(viewport->buffer_id != 0);
	samps_per_sec = viewport->samples_per_second;
	cache = GDAM_PRODUCER_CACHE(skin->model);
	g_return_if_fail(cache != NULL);
	cache->buffer_id = viewport->buffer_id;
	if (viewport->has_selection) {
		cache->start = (int)(viewport->selection_start * samps_per_sec);
		cache->length = (int)(viewport->selection_length*samps_per_sec);
	} else {
		cache->start = (int)(viewport->time_start * samps_per_sec);
		cache->length = (int)(viewport->time_length * samps_per_sec);
	}
	gdam_producer_play(GDAM_PRODUCER(cache), NULL);
}

static gboolean gdam_buffer_editor_play_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
                                  GTK_SIGNAL_FUNC(play_callback),
				  (GtkObject*) skin);
	return TRUE;
}

typedef struct _CutInfo CutInfo;
struct _CutInfo {
	char* description;
	GdamChannel* channel;
	GdamBufferViewport* viewport;
};

static void successfully_cut(int num_args, GdamArg* args, gpointer user_data)
{
	int buffer_id = 0;
	GdamBufferManagerInfo* info;
	CutInfo* cut_info = user_data;
	GdamBufferViewport* viewport = cut_info->viewport;
	int i;
	for (i = 0; i < num_args; i++) {
		if (strcmp(args[i].name, "buffer_id") == 0
		 && args[i].type == GDAM_ARG_TYPE_INT) {
			buffer_id = GDAM_ARG_INT(args+i);
			continue;
		}
		gdam_debug("got unknown arg %s as response to buffer_cut",
			args[i].name);
	}
	if (buffer_id == 0) {
		g_warning("cut/copy was successful - but didn't get buffer_id");
		g_free(cut_info->description);
		g_free(cut_info);
		return;
	}
	info = g_new(GdamBufferManagerInfo, 1);
	info->buffer_id = buffer_id;
	info->channel = cut_info->channel;
	info->description = cut_info->description;
	info->length = -1;
	gdam_buffer_manager_add_buffer(info);
	g_free(cut_info);
	gdam_buffer_manager_refresh(buffer_id);
	gdam_buffer_viewport_buffer_changed(viewport);
}
static void successfully_copied(int num_args, GdamArg* args, gpointer user_data)
{
	int buffer_id = 0;
	GdamBufferManagerInfo* info;
	CutInfo* cut_info = user_data;
	int i;
	for (i = 0; i < num_args; i++) {
		if (strcmp(args[i].name, "buffer_id") == 0
		 && args[i].type == GDAM_ARG_TYPE_INT) {
			buffer_id = GDAM_ARG_INT(args+i);
			continue;
		}
		gdam_debug("got unknown arg %s as response to buffer_cut",
			args[i].name);
	}
	if (buffer_id == 0) {
		g_warning("copy was successful - but didn't get buffer_id");
		g_free(cut_info->description);
		g_free(cut_info);
		return;
	}
	info = g_new(GdamBufferManagerInfo, 1);
	info->buffer_id = buffer_id;
	info->channel = cut_info->channel;
	info->description = cut_info->description;
	info->length = -1;
	gdam_buffer_manager_add_buffer(info);
	g_free(cut_info);
}
static void failed_to_cut(gpointer user_data)
{
	CutInfo* cut_info = user_data;
	g_free(cut_info->description);
	g_free(cut_info);
}
typedef struct _CutCallbackInfo CutCallbackInfo;
struct _CutCallbackInfo {
	gboolean	      make_buffer;
	GdamBufferEditorSkin* editor_skin;
};

static void cut_callback(CutCallbackInfo* callback_info)
{
	int start, length;
	GdamBufferEditorSkin* editor_skin = callback_info->editor_skin;
	GdamBufferViewport* viewport = editor_skin->viewport;
	int samps_per_sec;
	GdamArg* args;
	CutInfo* cut_info;
	char* description;
	GdamRequest* request;
	g_return_if_fail(viewport != NULL);
	g_return_if_fail(viewport->buffer_id != 0);
	samps_per_sec = viewport->samples_per_second;
	if (!viewport->has_selection) {
		return;
	}
	start = (int)(viewport->selection_start * samps_per_sec);
	length = (int)(viewport->selection_length * samps_per_sec);
	args = gdam_arg_new_from_valist(4,
	                                "start",
					GDAM_ARG_TYPE_INT,
					start,
	                                "length",
					GDAM_ARG_TYPE_INT,
					length,
	                                "make_buffer",
					GDAM_ARG_TYPE_INT,
					callback_info->make_buffer,
	                                "cut",
					GDAM_ARG_TYPE_INT,
					1);
	if (length == -1)
		description = g_strdup_printf("cut [%4.3g-] of buffer %d",
					(gdouble)start / samps_per_sec,
					viewport->buffer_id);
	else
		description = g_strdup_printf("cut [%4.3g-%4.3g] of buffer %d",
					(gdouble)start / samps_per_sec,
					(gdouble)(start+length) / samps_per_sec,
					viewport->buffer_id);
	cut_info = g_new(CutInfo, 1);
	cut_info->description = description;
	cut_info->channel = GDAM_SKIN(viewport)->channel;
	cut_info->viewport = viewport;
	request = gdam_request_call_function
	                          (GDAM_PROTOCOL_BUFFER_ARG,
                                   viewport->buffer_id,
	                           "process_region",
				   4,
				   args,
				   successfully_cut,
				   failed_to_cut,
				   cut_info);
	gdam_channel_append_request(cut_info->channel, request);
	gdam_arg_free_array(4, args);
	gdam_buffer_viewport_buffer_changed(viewport);
}

static gboolean gdam_buffer_editor_cut_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	CutCallbackInfo* callback_info = g_new(CutCallbackInfo, 1);
	callback_info->make_buffer = TRUE;
	callback_info->editor_skin = GDAM_BUFFER_EDITOR_SKIN(skin);
	gtk_signal_connect_full(object,
	                        signal_name,
                                GTK_SIGNAL_FUNC(cut_callback),
				NULL,
				callback_info,
				g_free,
				TRUE,
				FALSE);
	return TRUE;
}
static gboolean gdam_buffer_editor_kill_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	CutCallbackInfo* callback_info = g_new(CutCallbackInfo, 1);
	callback_info->make_buffer = FALSE;
	callback_info->editor_skin = GDAM_BUFFER_EDITOR_SKIN(skin);
	gtk_signal_connect_full(object,
	                        signal_name,
                                GTK_SIGNAL_FUNC(cut_callback),
				NULL,
				callback_info,
				g_free,
				TRUE,
				FALSE);
	return TRUE;
}


static void copy_callback(GdamBufferEditorSkin* editor_skin)
{
	int start, length;
	int samps_per_sec;
	GdamBufferViewport* viewport = editor_skin->viewport;
	GdamArg* args;
	CutInfo* cut_info;
	char* description;
	GdamRequest* request;
	g_return_if_fail(viewport != NULL);
	g_return_if_fail(viewport->buffer_id != 0);
	if (!viewport->has_selection) {
		return;
	}
	samps_per_sec = viewport->samples_per_second;
	start = (int)(viewport->selection_start * samps_per_sec);
	length = (int)(viewport->selection_length * samps_per_sec);
	args = gdam_arg_new_from_valist(4,
	                                "start",
					GDAM_ARG_TYPE_INT,
					start,
	                                "length",
					GDAM_ARG_TYPE_INT,
					length,
	                                "make_buffer",
					GDAM_ARG_TYPE_INT,
					1,
	                                "cut",
					GDAM_ARG_TYPE_INT,
					0);
	if (length == -1)
		description = g_strdup_printf("copy [%4.3g-] of buffer %d",
					(gdouble)start / samps_per_sec,
					viewport->buffer_id);
	else
		description = g_strdup_printf("copy [%4.3g-%4.3g] of buffer %d",
					(gdouble)start / samps_per_sec,
					(gdouble)(start+length) / samps_per_sec,
					viewport->buffer_id);
	cut_info = g_new(CutInfo, 1);
	cut_info->description = description;
	cut_info->channel = GDAM_SKIN(viewport)->channel;
	cut_info->viewport = viewport;
	request = gdam_request_call_function
	                          (GDAM_PROTOCOL_BUFFER_ARG,
                                   viewport->buffer_id,
	                           "process_region",
				   4,
				   args,
				   successfully_copied,
				   failed_to_cut,
				   cut_info);
	gdam_channel_append_request(cut_info->channel, request);
	gdam_arg_free_array(4, args);
}

static gboolean gdam_buffer_editor_copy_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
                                  GTK_SIGNAL_FUNC(copy_callback),
				  (GtkObject*) skin);
	return TRUE;
}

static void
successfully_replaced(int num_arg, GdamArg* args, gpointer user_data)
{
	GdamBufferViewport* viewport = GDAM_BUFFER_VIEWPORT(user_data);
	if (viewport) {
		if (viewport->buffer_id)
			gdam_buffer_manager_refresh(viewport->buffer_id);
		gdam_buffer_viewport_buffer_changed(viewport);
	}
	gtk_object_unref(GTK_OBJECT(user_data));
}

static void
failed_to_replaced(gpointer user_data)
{
	gtk_object_unref(GTK_OBJECT(user_data));
}
static void replace_callback(GdamBufferEditorSkin* editor_skin)
{
	int start, length;
	int buffer_id;
	int samps_per_sec;
	GdamBufferViewport* viewport = editor_skin->viewport;
	GdamArg* args;
	GdamRequest* request;
	g_return_if_fail(viewport != NULL);
	g_return_if_fail(viewport->buffer_id != 0);
	if (!viewport->has_selection) {
		return;
	}
	samps_per_sec = viewport->samples_per_second;
	start = (int)(viewport->selection_start * samps_per_sec);
	length = (int)(viewport->selection_length * samps_per_sec);
	buffer_id = gdam_buffer_manager_obtain_id(NULL);
	gdam_buffer_manager_unref(buffer_id, NULL);
	args = gdam_arg_new_from_valist(3,
	                                "start",
					GDAM_ARG_TYPE_INT,
					start,
	                                "length",
					GDAM_ARG_TYPE_INT,
					start,
	                                "insert_buffer_id",
					GDAM_ARG_TYPE_INT,
					buffer_id);
	gtk_object_ref(GTK_OBJECT(viewport));
	request = gdam_request_call_function
	                          (GDAM_PROTOCOL_BUFFER_ARG,
                                   viewport->buffer_id,
	                           "replace_region",
				   3,
				   args,
				   successfully_replaced,
				   failed_to_replaced,
				   viewport);
	gdam_channel_append_request(GDAM_SKIN(editor_skin)->channel, request);
	gdam_arg_free_array(3, args);
}
static gboolean gdam_buffer_editor_replace_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
                                  GTK_SIGNAL_FUNC(replace_callback),
				  (GtkObject*) skin);
	return TRUE;
}


typedef struct _InsertInfo InsertInfo;
struct _InsertInfo {
	int			multiplicity;
	gboolean		after_selection;
	GdamBufferEditorSkin*   editor_skin;
};

static void
successfully_inserted(int num_arg, GdamArg* args, gpointer user_data)
{
	InsertInfo* insert_info = user_data;
	GdamBufferEditorSkin* editor_skin = insert_info->editor_skin;
	GdamBufferViewport* viewport = editor_skin->viewport;
	if (viewport) {
		if (viewport->buffer_id)
			gdam_buffer_manager_refresh(viewport->buffer_id);
		gdam_buffer_viewport_buffer_changed(viewport);
	}
	gtk_object_unref(GTK_OBJECT(viewport));
}

static void
failed_to_insert(gpointer user_data)
{
	InsertInfo* insert_info = user_data;
	GdamBufferEditorSkin* editor_skin = insert_info->editor_skin;
	GdamBufferViewport* viewport = editor_skin->viewport;
	gtk_object_unref(GTK_OBJECT(viewport));
}
static void insert_callback(InsertInfo* insert_info)
{
	int start;
	int buffer_id;
	int samps_per_sec;
	GdamBufferEditorSkin* editor_skin = insert_info->editor_skin;
	GdamBufferViewport* viewport = editor_skin->viewport;
	GdamArg* args;
	GdamRequest* request;
	g_return_if_fail(viewport != NULL);
	g_return_if_fail(viewport->buffer_id != 0);
	if (!viewport->has_selection) {
		return;
	}
	samps_per_sec = viewport->samples_per_second;
	if (insert_info->after_selection)
		start = (int)(samps_per_sec * (viewport->selection_start
				            + viewport->selection_length));
	else
		start = (int)(viewport->selection_start * samps_per_sec);
	buffer_id = gdam_buffer_manager_obtain_id(NULL);
	gdam_buffer_manager_unref(buffer_id, NULL);
	args = gdam_arg_new_from_valist(3,
	                                "start",
					GDAM_ARG_TYPE_INT,
					start,
	                                "multiplicity",
					GDAM_ARG_TYPE_INT,
					insert_info->multiplicity,
	                                "insert_buffer_id",
					GDAM_ARG_TYPE_INT,
					buffer_id);
	gtk_object_ref(GTK_OBJECT(viewport));
	request = gdam_request_call_function
	                          (GDAM_PROTOCOL_BUFFER_ARG,
                                   viewport->buffer_id,
	                           "insert_region",
				   3,
				   args,
				   successfully_inserted,
				   failed_to_insert,
				   insert_info);
	gdam_channel_append_request(GDAM_SKIN(editor_skin)->channel, request);
	gdam_arg_free_array(3, args);
}
static gboolean gdam_buffer_editor_insert_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	InsertInfo* insert_info = g_new(InsertInfo, 1);
	int i;
	insert_info->editor_skin = GDAM_BUFFER_EDITOR_SKIN(skin);
	insert_info->multiplicity = 1;
	insert_info->after_selection = FALSE;

	for (i = 1; i < num_handler_names; i++) {
		char* name = handler_names[i];
		if (strncmp(name, "times=", 6) == 0) {
			int times = atoi(name + 6);
			if (times > 0)
				insert_info->multiplicity = times;
			continue;
		}
		if (strcmp(name, "after") == 0) {
			insert_info->after_selection = TRUE;
			continue;
		}
		g_message("unknown handler name for buffer_insert: %s", name);
	}
	
	gtk_signal_connect_full(object,
	                        signal_name,
                                GTK_SIGNAL_FUNC(insert_callback),
				NULL,
				insert_info,
				g_free,
				TRUE,
				FALSE);
	return TRUE;
}

/*
 *                                    _
 *  _ __ _____   _____ _ __ ___  __ _| |
 * | '__/ _ \ \ / / _ \ '__/ __|/ _` | |
 * | | |  __/\ V /  __/ |  \__ \ (_| | |
 * |_|  \___| \_/ \___|_|  |___/\__,_|_|
 * 
 */
static void
successfully_reversed(int num_arg, GdamArg* args, gpointer user_data)
{
	GdamBufferViewport* viewport = GDAM_BUFFER_VIEWPORT(user_data);
	gdam_buffer_viewport_buffer_changed(viewport);
	gtk_object_unref(GTK_OBJECT(viewport));
}

static void
failed_to_reverse(gpointer user_data)
{
	GdamBufferViewport* viewport = GDAM_BUFFER_VIEWPORT(user_data);
	gtk_object_unref(GTK_OBJECT(viewport));
}

static void reverse_callback(GdamBufferEditorSkin* editor_skin)
{
	GdamModel* model = GDAM_SKIN(editor_skin)->model;
	guint buffer_id;
	GdamArg* args;
	GdamRequest* request;
	GdamBufferViewport* viewport;
	int start, length;
	int samps_per_sec;

	g_return_if_fail(model != NULL);
	viewport = editor_skin->viewport;
	g_return_if_fail(viewport != NULL);
	buffer_id = editor_skin->viewport->buffer_id;
	if (buffer_id == 0)
		return;

	samps_per_sec = viewport->samples_per_second;
	start = (int)(samps_per_sec * viewport->selection_start);
	length = (int)(samps_per_sec * viewport->selection_length);
	args = gdam_arg_new_from_valist(2,
	                                "start",
					GDAM_ARG_TYPE_INT,
					start,
	                                "length",
					GDAM_ARG_TYPE_INT,
					length);
	gtk_object_ref(GTK_OBJECT(viewport));
	request = gdam_request_call_function
	                          (GDAM_PROTOCOL_BUFFER_ARG,
                                   buffer_id,
	                           "reverse_region",
				   2,
				   args,
				   successfully_reversed,
				   failed_to_reverse,
				   editor_skin->viewport);
	gdam_channel_append_request(GDAM_SKIN(editor_skin)->channel, request);
	gdam_arg_free_array(2, args);
}

static gboolean gdam_buffer_editor_reverse_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
                                  GTK_SIGNAL_FUNC(reverse_callback),
				  (GtkObject*) skin);
	return TRUE;
}

static void select_all_callback(GdamBufferEditorSkin* editor_skin)
{
	GdamBufferViewport* viewport = editor_skin->viewport;
	g_return_if_fail(viewport != NULL);
	viewport->selection_start = viewport->time_start;
	viewport->selection_length = viewport->time_length;
	gdam_buffer_viewport_queue_redraw(viewport);
}

static gboolean gdam_buffer_editor_select_all_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	gtk_signal_connect_object(object,
	                          signal_name,
                                  GTK_SIGNAL_FUNC(select_all_callback),
				  (GtkObject*) skin);
	return TRUE;
}

static void update_start_time_label(GdamBufferViewport* view,
                                    GtkLabel*           to_change)
{
	char buf[1024];
	snprintf(buf, sizeof(buf), "%4.4gs", view->time_start);
	gtk_label_set_text(to_change, buf);
}

static gboolean gdam_buffer_editor_start_time_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamBufferEditorSkin* editor_skin = GDAM_BUFFER_EDITOR_SKIN(skin);
	if (! GTK_IS_LABEL(object)) {
		g_message("start_time_label can only be attached to a label");
		return TRUE;
	}
	if (editor_skin->viewport == NULL)
		return FALSE;
	gtk_signal_connect_while_alive
			(GTK_OBJECT(editor_skin->viewport),
			 "view_changed",
			 GTK_SIGNAL_FUNC(update_start_time_label),
			 object,
			 object);
	return TRUE;
}


static void update_end_time_label(GdamBufferViewport* view,
                                    GtkLabel*           to_change)
{
	char buf[1024];
	gdouble end = view->time_start + view->time_length;
	snprintf(buf, sizeof(buf), "%4.4gs", end);
	gtk_label_set_text(to_change, buf);
}

static gboolean gdam_buffer_editor_end_time_label_handler
				(GdamSkin*          skin,
				 GtkObject*         object,
				 const char*        signal_name,
				 int                num_handler_names,
				 char**             handler_names)
{
	GdamBufferEditorSkin* editor_skin = GDAM_BUFFER_EDITOR_SKIN(skin);
	if (! GTK_IS_LABEL(object)) {
		g_message("end_time_label can only be attached to a label");
		return TRUE;
	}
	if (editor_skin->viewport == NULL)
		return FALSE;
	gtk_signal_connect_while_alive
			(GTK_OBJECT(editor_skin->viewport),
			 "view_changed",
			 GTK_SIGNAL_FUNC(update_end_time_label),
			 object,
			 object);
	return TRUE;
}

static GdamSkinGenericConnect connect_infos[] =
{
	{ "gdam_buffer_editor_viewport",
           gdam_buffer_editor_viewport_handler			},
	{ "gdam_buffer_editor_log_zoom_adjustment",
           gdam_buffer_editor_log_zoom_adjustment_handler	},
	{ "gdam_buffer_editor_center_adjustment",
           gdam_buffer_editor_center_adjustment_handler		},
	{ "gdam_buffer_editor_buffer_id_dropped",
           gdam_buffer_editor_buffer_id_dropped_handler		},
	{ "gdam_buffer_editor_zoom_to_selection",
	   gdam_buffer_editor_zoom_to_selection_handler         },
	{ "gdam_buffer_editor_zoom_to_sample",
	   gdam_buffer_editor_zoom_to_sample_handler            },
	{ "gdam_buffer_editor_set_default",
           gdam_buffer_editor_set_default_handler		},
        { "gdam_buffer_editor_play",
           gdam_buffer_editor_play_handler                      },
        { "gdam_buffer_editor_cut",
           gdam_buffer_editor_cut_handler			},
        { "gdam_buffer_editor_kill",
           gdam_buffer_editor_kill_handler			},
        { "gdam_buffer_editor_copy",
           gdam_buffer_editor_copy_handler			},
        { "gdam_buffer_editor_replace",
           gdam_buffer_editor_replace_handler			},
        { "gdam_buffer_editor_insert",
           gdam_buffer_editor_insert_handler			},
        { "gdam_buffer_editor_reverse",
           gdam_buffer_editor_reverse_handler                   },
	{ "gdam_buffer_editor_start_time_label",
	   gdam_buffer_editor_start_time_label_handler		},
	{ "gdam_buffer_editor_end_time_label",
	   gdam_buffer_editor_end_time_label_handler		},
        { "gdam_buffer_editor_select_all",
           gdam_buffer_editor_select_all_handler                }
};
static void gdam_buffer_editor_skin_init(GdamBufferEditorSkin* editor)
{
}
static void gdam_buffer_editor_skin_class_init(GdamSkinClass* skin_class)
{
	skin_class->build_new_model =
				gdam_buffer_editor_build_new_model;
	gdam_skin_class_register_generic_handlers(skin_class,
				G_N_ELEMENTS(connect_infos),
				connect_infos);
}


static gboolean       gdam_buffer_editor_build_new_model
                                                 (GdamSkin*      skin,
                                                  GdamChannel*   channel,
						  int		 num_options,
						  char**         options)
{
	GdamModel* model;
	GdamBufferEditorSkin* editor = GDAM_BUFFER_EDITOR_SKIN(skin);
	GdamBufferViewport* viewport;
	model = GDAM_MODEL(gtk_type_new(GDAM_TYPE_PRODUCER_CACHE));
	model->channel = channel;
	skin->model = model;
	viewport = gdam_buffer_viewport_new_shown(channel);
	gtk_object_ref(GTK_OBJECT(viewport));
	gtk_object_sink(GTK_OBJECT(viewport));
	editor->viewport = viewport;

	return TRUE;
}

GtkType gdam_buffer_editor_skin_get_type()
{
	static GtkType buffer_editor_skin_type = 0;
	if (!buffer_editor_skin_type) {
		static const GtkTypeInfo buffer_editor_skin_info = {
			"GdamBufferEditorSkin",
			sizeof(GdamBufferEditorSkin),
			sizeof(GdamBufferEditorSkinClass),
			(GtkClassInitFunc) gdam_buffer_editor_skin_class_init,
			(GtkObjectInitFunc) gdam_buffer_editor_skin_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		GtkType parent = GDAM_TYPE_SKIN;
		buffer_editor_skin_type = gtk_type_unique(parent,
		                                 &buffer_editor_skin_info);
	}
	return buffer_editor_skin_type;
}
